from geopandas.array import GeometryDtype
from pandera.pandas import Check, Column, DataFrameSchema

from geost import config

# Point header schema
pointheader = DataFrameSchema(
    columns={
        "nr": Column(str, unique=True),
        "x": Column(float),
        "y": Column(float),
        "surface": Column(float),
        "end": Column(float),
        "geometry": Column(GeometryDtype()),
    },
    checks=Check(lambda df: df["end"] < df["surface"], element_wise=False),
    coerce=True,
    strict=False,
    drop_invalid_rows=config.validation.DROP_INVALID,
    name="Point header",
    description="Schema for validating point header data",
)

# Line header schema
lineheader = DataFrameSchema(
    columns={
        "nr": Column(str, unique=True),
        "geometry": Column(GeometryDtype(), nullable=False),
    },
    coerce=True,
    strict=False,
    drop_invalid_rows=config.validation.DROP_INVALID,
    name="Line header",
    description="Schema for validating line header data",
)

# Layer data schema non-inclined
layerdata = DataFrameSchema(
    columns={
        "nr": Column(str),
        "x": Column(float),
        "y": Column(float),
        "surface": Column(float),
        "end": Column(float),
        "top": Column(float),
        "bottom": Column(float),
    },
    checks=[
        Check(lambda df: df["bottom"] >= df["top"], element_wise=False),
        Check(lambda df: df["end"] < df["surface"], element_wise=False),
    ],
    coerce=True,
    strict=False,
    drop_invalid_rows=config.validation.DROP_INVALID,
    name="Layer data non-inclined",
    description="Schema for validating generic layer data that is not inclined",
)

# Layer data schema inclined
layerdata_inclined = DataFrameSchema(
    columns={
        "nr": Column(str),
        "x": Column(float),
        "y": Column(float),
        "x_bot": Column(float),
        "y_bot": Column(float),
        "surface": Column(float),
        "end": Column(float),
        "top": Column(float),
        "bottom": Column(float),
    },
    checks=[
        Check(lambda df: df["bottom"] >= df["top"], element_wise=False),
        Check(lambda df: df["end"] < df["surface"], element_wise=False),
    ],
    coerce=True,
    strict=False,
    drop_invalid_rows=config.validation.DROP_INVALID,
    name="Layer data inclined",
    description="Schema for validating generic layer data that is inclined",
)

# Discrete data schema non-inclined
discretedata = DataFrameSchema(
    columns={
        "nr": Column(str),
        "x": Column(float),
        "y": Column(float),
        "surface": Column(float),
        "end": Column(float),
        "depth": Column(float),
    },
    checks=[
        Check(lambda df: df["depth"] >= 0, element_wise=False),
        Check(lambda df: df["end"] < df["surface"], element_wise=False),
    ],
    coerce=True,
    strict=False,
    drop_invalid_rows=config.validation.DROP_INVALID,
    name="Discrete data non-inclined",
    description="Schema for validating generic discrete data that is not inclined",
)

# Discrete data schema inclined
discretedata_inclined = DataFrameSchema(
    columns={
        "nr": Column(str),
        "x": Column(float),
        "y": Column(float),
        "x_bot": Column(float),
        "y_bot": Column(float),
        "surface": Column(float),
        "end": Column(float),
        "depth": Column(float),
    },
    checks=[
        Check(lambda df: df["depth"] >= 0, element_wise=False),
        Check(lambda df: df["end"] < df["surface"], element_wise=False),
    ],
    coerce=True,
    strict=False,
    drop_invalid_rows=config.validation.DROP_INVALID,
    name="Discrete data inclined",
    description="Schema for validating generic discrete data that is inclined",
)

# Grain size data schema
grainsize_data = DataFrameSchema(
    columns={
        "nr": Column(str),
        "sample_nr": Column(str),
        "x": Column(float),
        "y": Column(float),
        "top": Column(float),
        "bottom": Column(float),
        "d_low": Column(float),
        "d_high": Column(float),
        "percentage": Column(float),
    },
    checks=[
        Check(lambda df: df["bottom"] >= df["top"], element_wise=False),
        Check(lambda df: df["d_high"] > df["d_low"], element_wise=False),
        Check(
            lambda df: (df["percentage"] >= 0) & (df["percentage"] <= 100),
            element_wise=False,
        ),
    ],
    coerce=True,
    strict=False,
    drop_invalid_rows=config.validation.DROP_INVALID,
    name="Grain size data",
    description="Schema for validating grain size data",
)



from pathlib import Path
from typing import Literal, NamedTuple

import numpy as np
import pytest
from numpy.testing import assert_array_almost_equal

from geost.io.parsers import CptGefFile
from geost.io.parsers.gef_parsers import ColumnInfo


def dummy_cpt_data():
    z = -0.5
    columns = ["length", "qc", "fs", "rf"]
    columninfo = {ii: ColumnInfo(c, None, None, None) for ii, c in enumerate(columns)}
    columnvoid = {0: -9999.0, 1: -9999.0, 2: -999.0, 3: -9999.0}

    length = ["1.300000", "1.320000", "1.340000", "1.360000", "1.380000"]
    qc = ["-9999.0000", "0.227523", "0.279521", "0.327816", "-9999.0000"]
    fs = ["-999.000", "0.010076", "0.014140", "-999.000", "0.020937"]
    rf = ["-9999.0000", "3.561200", "4.462611", "5.452729", "5.697089"]

    data = np.array([length, qc, fs, rf]).T

    return z, columninfo, columnvoid, data


class CptInfo(NamedTuple):
    nr: str
    x: float
    y: float
    z: float
    end: float
    ncols: int
    system: str
    reference: str
    error: float
    nrecords: int


class TestCptGefParser:
    @pytest.fixture
    def test_cpt_files(self, testdatadir):
        test_cpt_files = list(testdatadir.glob("gef/*.gef"))
        return test_cpt_files

    @pytest.fixture
    def cpt_a(self, testdatadir):
        cpt = CptGefFile(testdatadir / r"gef/83268_DKMP003_wiertsema.gef")
        info_to_test = CptInfo(
            "DKMP_D03",
            176161.1,
            557162.1,
            -0.06,
            -0.06 - 37.317295,
            11,
            "31000",
            "NAP",
            0.01,
            1815,
        )
        return cpt, info_to_test

    @pytest.fixture
    def cpt_b(self, testdatadir):
        cpt = CptGefFile(testdatadir / r"gef/AZZ158_gem_rotterdam.gef")
        info_to_test = CptInfo(
            "AZZ158",
            0.0,
            0.0,
            5.05,
            5.05 - 59.5,
            6,
            "0",
            "NAP",
            None,
            2980,
        )
        return cpt, info_to_test

    @pytest.fixture
    def cpt_c(self, testdatadir):
        cpt = CptGefFile(testdatadir / r"gef/CPT000000157983_IMBRO.gef")
        info_to_test = CptInfo(
            "CPT000000157983",
            176416.1,
            557021.9,
            -5.5,
            -5.5 - 28.84,
            9,
            "28992",
            "NAP",
            None,
            1310,
        )
        return cpt, info_to_test

    @pytest.fixture
    def cpt_d(self, testdatadir):
        cpt = CptGefFile(testdatadir / r"gef/CPT10.gef")
        info_to_test = CptInfo(
            "YANGTZEHAVEN CPT 10",
            61949.0,
            443624.0,
            -17.69,
            -17.69 - 5.75,
            4,
            "31000",
            "NAP",
            None,
            207,
        )
        return cpt, info_to_test

    @pytest.fixture
    def dummy_cpt_with_rf(self):
        z, columninfo, columnvoid, data = dummy_cpt_data()

        cpt = CptGefFile()
        cpt.z = z
        cpt.columninfo = columninfo
        cpt.columnvoid = columnvoid
        cpt._data = data

        return cpt

    @pytest.mark.unittest
    def test_read_files(self, test_cpt_files: list[Path]):
        for f in test_cpt_files:
            cpt = CptGefFile(f)
            assert isinstance(cpt, CptGefFile)

    @pytest.mark.integrationtest
    @pytest.mark.parametrize("cpt_test", ["cpt_a", "cpt_b", "cpt_c", "cpt_d"])
    def test_cpt_parsing_result(
        self,
        cpt_test: (
            Literal["cpt_a"] | Literal["cpt_b"] | Literal["cpt_c"] | Literal["cpt_d"]
        ),
        request: pytest.FixtureRequest,
    ):
        cpt, test_info = request.getfixturevalue(cpt_test)

        assert cpt.nr == test_info.nr
        assert cpt.x == test_info.x
        assert cpt.y == test_info.y
        assert cpt.z == test_info.z
        assert cpt.enddepth == test_info.end
        assert cpt.ncolumns == test_info.ncols
        assert cpt.coord_system == test_info.system
        assert cpt.reference_system == test_info.reference
        assert cpt.delta_z == test_info.error

        critical_cols_from_file = ["length", "qc", "fs"]
        assert all(col in cpt.columns for col in critical_cols_from_file)
        assert len(cpt.df) == test_info.nrecords

    @pytest.mark.unittest
    def test_to_dataframe(self, dummy_cpt_with_rf: CptGefFile):
        dummy_cpt_with_rf.to_df()

        target_columns = ["length", "qc", "fs", "rf", "depth"]
        target_depth = [1.30, 1.32, 1.34, 1.36, 1.38]

        df = dummy_cpt_with_rf._df
        nancount = np.sum(np.isnan(df.values))

        assert all(df.dtypes == "float64")
        assert nancount == 5
        assert all(df.columns == target_columns)

        assert_array_almost_equal(df["depth"], target_depth)



from geost.io.xml import resolvers

BRO = {
    "payload_root": "dispatchDocument",
    "nr": {"xpath": "brocom:broId"},
    "location": {
        "xpath": "deliveredLocation/bhrgcom:location/gml:Point/gml:pos",
        "resolver": resolvers.parse_coordinates,
        "el-attr": "text",
    },
    "crs": {
        "xpath": "deliveredLocation/bhrgcom:location/gml:Point",
        "resolver": resolvers.parse_crs,
    },
    "surface": {
        "xpath": "deliveredVerticalPosition/bhrgcom:offset",
        "resolver": resolvers.safe_float,
        "el-attr": "text",
    },
    "vertical_datum": {
        "xpath": "deliveredVerticalPosition/bhrgcom:verticalDatum",
        "el-attr": "text",
    },
    "end": {
        "xpath": "boring/bhrgcom:Boring/bhrgcom:finalDepthBoring",
        "resolver": resolvers.safe_float,
        "el-attr": "text",
    },
    "data": {
        "xpath": "boreholeSampleDescription/bhrgcom:BoreholeSampleDescription/bhrgcom:descriptiveBoreholeLog/bhrgcom:DescriptiveBoreholeLog",
        "resolver": resolvers.process_bhrgt_data,
        "layer-attributes": ["upperBoundary", "lowerBoundary", "soilNameNEN5104"],
    },
}

schemas = {"BRO": BRO}



from pyproj import CRS, Transformer
from pyproj.crs import CompoundCRS


def horizontal_reference_transformer(
    epsg_from: str | int | CRS, epsg_to: str | int | CRS
):
    transformer = Transformer.from_crs(
        CRS(epsg_from), CRS(epsg_to), always_xy=True, accuracy=1.0
    )
    return transformer


def vertical_reference_transformer(
    horizontal_epsg: str | int | CRS,
    epsg_from: str | int | CRS,
    epsg_to: str | int | CRS,
):
    epsg_from = CompoundCRS("Temporary source CRS", [horizontal_epsg, epsg_from])
    epsg_to = CompoundCRS("Temporary target CRS", [horizontal_epsg, epsg_to])

    if epsg_from.is_vertical and epsg_to.is_vertical:
        transformer = Transformer.from_crs(epsg_from, epsg_to, always_xy=True)
    else:
        raise TypeError(
            "Given EPSG numbers must be known vertical datums, see spatialreference.org"
        )
    return transformer



"""
The basic XML parsing design (i.e. `_read_xml`) in this module is based on Pygef's approach
but the aim of this module is to provide a more flexible and extensible approach to parsing
XML files.

See: https://github.com/cemsbv/pygef/tree/master/src/pygef/broxml for the original design.

"""

import io
import warnings
from pathlib import Path
from typing import Any, Callable, Iterable

import pandas as pd
from lxml import etree

from . import schemas

_BaseParser = etree.XMLParser(
    resolve_entities=False, dtd_validation=False
)  # Speeds up parsing by not resolving entities or validating DTDs.


def read(
    files: str | Path | io.StringIO | Iterable[str | Path],
    reader: Callable,
    **kwargs: dict[str, Any],
) -> tuple[pd.DataFrame, pd.DataFrame]:
    """
    Read one or multiple XML files or string objects and create `geost` compatible separate
    `header` and `data` DataFrames.

    Parameters
    ----------
    files : str | Path | io.StringIO | Iterable[str | Path]
        List of XML files to process.
    reader : Callable
        Function to read the XML data. See `geost.io.xml.READERS` for available XML readers.
    **kwargs : dict[str, Any]
        Additional keyword arguments to pass to the reader function.

    Returns
    -------
    tuple of pandas.DataFrame
        Tuple containing the header DataFrame and the data DataFrame.

    """
    if isinstance(files, (str, Path, io.StringIO)):
        files = [files]

    header = []
    data = []
    lengths = []

    for file in files:
        xml_data = reader(file, **kwargs)
        xml_data["x"], xml_data["y"] = xml_data.pop("location")
        df = pd.DataFrame(xml_data.pop("data"))

        header.append(xml_data)
        data.append(df)
        lengths.append(len(df))

    header = pd.DataFrame(header)
    header["end"] = header["surface"] - header["end"]
    data = pd.concat(data, ignore_index=True)

    # Add relevant header attributes to the data DataFrame
    attributes_for_data = ["nr", "x", "y", "surface", "end"]
    header_data = header.loc[header.index.repeat(lengths), attributes_for_data]
    data = pd.concat([header_data.reset_index(drop=True), data], axis=1)

    if "upperBoundary" in data.columns or "lowerBoundary" in data.columns:
        data.rename(
            columns={"upperBoundary": "top", "lowerBoundary": "bottom"}, inplace=True
        )
        data[["top", "bottom"]] = data[["top", "bottom"]].astype(float)

    return header, data


def _read_xml(
    file: bytes | str | Path,
    schema: dict[str, Any],
    payload_root: str = None,
    read_all: bool = False,
) -> dict | list[dict]:
    if isinstance(file, bytes):
        root = etree.fromstring(file, parser=_BaseParser)
    else:
        root = etree.parse(file, parser=_BaseParser).getroot()

    if payload_root is not None:
        xmldata = root.find(payload_root, root.nsmap)

        if xmldata is None:
            raise ValueError(f"Payload root '{payload_root}' not found in XML.")
    else:
        xmldata = root  # Try to parse the entire root using the given schema.

    output = []

    payloads = xmldata.findall("./*")
    for payload in payloads:
        data = {}

        for key, d in schema.items():
            if key == "payload_root":
                continue
            el = payload.find(d["xpath"], payload.nsmap)

            if el is not None:
                if "resolver" in d:
                    func = d["resolver"]

                    if "el-attr" in d:
                        el = getattr(el, d["el-attr"])

                    if attributes := d.get("layer-attributes", ""):
                        data[key] = func(el, attributes=attributes)
                    else:
                        data[key] = func(el)
                else:
                    data[key] = el.text
            else:
                data[key] = None

        if read_all:
            output.append(data)
        else:
            if len(payloads) > 1:
                warnings.warn(
                    "Multiple payloads found in XML but 'read_all' is set to False. "
                    "Set 'read_all' to True to read all payloads."
                )
            return data

    return output


def read_bhrgt(
    file: bytes | str | Path,
    company: str = None,
    schema: dict[str, Any] = None,
    read_all: bool = False,
) -> dict | list[dict]:
    """
    Read Geotechnical borehole data (BHR-GT) from an XML file or bytestring and extract
    relevant data based on a predefined or custom schema that describes the XML structure.

    Parameters
    ----------
    file : bytes | str | Path
        XML filepath-like or bytestring containing the XML data to read.
    company : str, optional
        Specify a company name to use a predefined schema for that company if available.
        See `bhrgt.SCHEMA` for available companies. The default is None, then company will
        default to "BRO" and the predefined BRO schema will be used if no custom schema
        is defined.
    schema : dict[str, Any], optional
        Custom schema used to parse the XML structure.
    read_all : bool, optional
        Generally, each XML file contains data for a single borehole but in some cases
        multiple boreholes can be present in the XML file. If set to True, all boreholes
        will be read from the XML file. The default is False, which reads only the first.
        A warning will be raised if multiple boreholes are found in the XML file and
        read_all is False.

    Returns
    -------
    dict | list[dict]
        Dict or list of dictionaries containing the extracted data from the XML file. If
        the XML file contains data for a single borehole, a dictionary will be returned.
        If multiple boreholes are found and `read_all` is True, a list of dictionaries
        will be returned.

    Raises
    ------
    ValueError
        Will raise a ValueError if no custom `schema` is defined and no predefined schema
        is available for the company.
    SyntaxError
        Raises a SyntaxError if the XML file does not conform to the expected schema. For
        example, with incorrect namespaces in the schema.

    Examples
    --------
    Read a BHR-GT XML for a specific company:
    >>> data = read_bhrgt("bhrgt_file.xml", company="Wiertsema") # Read for specific company

    Read a BHR-GT XML using a custom schema:
    >>> custom_schema = {
    ...     "payload_root": "dispatchDocument",
    ...     "bro_id": {"xpath": "brocom:broId"},
    ...     "date": {"xpath": "reportHistory/reportStartDate/brocom:date"}
    ...     "coordinates": {
    ...         "xpath": "./deliveredLocation/bhrgtcom:location/gml:Point/gml:pos",
    ...         "resolver": resolvers.parse_coordinates, # Add a custom function to parse coordinates
    ...         "el-attr": "text"  # Use the text attribute of the element
    ...     }
    ... }
    >>> data = read_bhrgt("bhrgt_file.xml", schema=custom_schema)  # Use custom schema

    Read a BHR-GT XML directly via a BRO API response:
    >>> from geost.bro import BroApi
    >>> api = BroApi()
    >>> response = api.get_objects("BHR000000449853", "BHR-GT") # Returns a list of XML strings
    >>> data = read_bhrgt(response[0])

    """
    company = company or "BRO"

    if schema is None:
        try:
            schema = schemas.bhrgt[company]
        except KeyError as e:
            raise ValueError(
                f"No predefined schema for '{company}' in BHR-GT. Supported companies are: "
                f"{schemas.bhrgt.keys()}. Define a custom schema or use a supported company."
            ) from e

    try:
        result = _read_xml(file, schema, schema.get("payload_root", None), read_all)
    except SyntaxError as e:
        raise SyntaxError(f"Invalid xml schema for XML file: {file}.") from e

    return result


def read_bhrp(
    file: bytes | str | Path,
    company: str = None,
    schema: dict[str, Any] = None,
    read_all: bool = False,
) -> dict | list[dict]:
    """
    Read borehole data (BHR-P) from an XML file or bytestring and extract relevant data
    based on a predefined or custom schema that describes the XML structure.

    Parameters
    ----------
    file : bytes | str | Path
        XML filepath-like or bytestring containing the XML data to read.
    company : str, optional
        Specify a company name to use a predefined schema for that company if available.
        See `bhrp.SCHEMA` for available companies. The default is None, then company will
        default to "BRO" and the predefined BRO schema will be used if no custom schema
        is defined.
    schema : dict[str, Any], optional
        Custom schema used to parse the XML structure.
    read_all : bool, optional
        Generally, each XML file contains data for a single borehole but in some cases
        multiple boreholes can be present in the XML file. If set to True, all boreholes
        will be read from the XML file. The default is False, which reads only the first.
        A warning will be raised if multiple boreholes are found in the XML file and
        read_all is False.

    Returns
    -------
    dict | list[dict]
        Dict or list of dictionaries containing the extracted data from the XML file. If
        the XML file contains data for a single borehole, a dictionary will be returned.
        If multiple boreholes are found and `read_all` is True, a list of dictionaries
        will be returned.

    Raises
    ------
    ValueError
        If no predefined schema is found for the specified company.
    SyntaxError
        If the XML file does not conform to the expected schema.

    """
    company = company or "BRO"

    if schema is None:
        try:
            schema = schemas.bhrp[company]
        except KeyError as e:
            raise ValueError(
                f"No predefined schema for '{company}' in BHR-P. Supported companies are: "
                f"{schemas.bhrp.keys()}. Define a custom schema or use a supported company."
            ) from e

    try:
        result = _read_xml(file, schema, schema.get("payload_root", None), read_all)
    except SyntaxError as e:
        raise SyntaxError(f"Invalid xml schema for XML file: {file}.") from e

    return result


def read_cpt(
    file: bytes | str | Path,
    company: str = None,
    schema: dict[str, Any] = None,
    read_all: bool = False,
) -> dict | list[dict]:
    """
    Read Cone Penetration Test (CPT) data from an XML file or bytestring and extract
    relevant data based on a predefined or custom schema that describes the XML structure.

    Parameters
    ----------
    file : bytes | str | Path
        XML filepath-like or bytestring containing the XML data to read.
    company : str, optional
        Specify a company name to use a predefined schema for that company if available.
        See `cpt.SCHEMA` for available companies. The default is None, then company will
        default to "BRO" and the predefined BRO schema will be used if no custom schema
        is defined.
    schema : dict[str, Any], optional
        Custom schema used to parse the XML structure.
    read_all : bool, optional
        Generally, each XML file contains data for a single CPT but in some cases
        multiple CPTs can be present in the XML file. If set to True, all CPTs will be
        read from the XML file. The default is False, which reads only the first.
        A warning will be raised if multiple CPTs are found in the XML file and
        read_all is False.

    Returns
    -------
    dict | list[dict]
        Dict or list of dictionaries containing the extracted data from the XML file. If
        the XML file contains data for a single CPT, a dictionary will be returned.
        If multiple CPTs are found and `read_all` is True, a list of dictionaries
        will be returned.

    Raises
    ------
    ValueError
        If no predefined schema is found for the specified company.
    SyntaxError
        If the XML file does not conform to the expected schema.

    """
    company = company or "BRO"

    if schema is None:
        try:
            schema = schemas.cpt[company]
        except KeyError as e:
            raise ValueError(
                f"No predefined schema for '{company}' in CPT. Supported companies are: "
                f"{schemas.cpt.keys()}. Define a custom schema or use a supported company."
            ) from e

    try:
        result = _read_xml(file, schema, schema.get("payload_root", None), read_all)
    except SyntaxError as e:
        raise SyntaxError(f"Invalid xml schema for XML file: {file}.") from e

    return result


def read_bhrg(
    file: bytes | str | Path,
    company: str = None,
    schema: dict[str, Any] = None,
    read_all: bool = False,
) -> dict | list[dict]:
    """
    Read Geological borehole data (BHR-G) from an XML file or bytestring and extract
    relevant data based on a predefined or custom schema that describes the XML structure.

    Parameters
    ----------
    file : bytes | str | Path
        XML filepath-like or bytestring containing the XML data to read.
    company : str, optional
        Specify a company name to use a predefined schema for that company if available.
        See `schemas.bhrg` for available companies. The default is None, then company will
        default to "BRO" and the predefined BRO schema will be used if no custom schema
        is defined.
    schema : dict[str, Any], optional
        Custom schema used to parse the XML structure.
    read_all : bool, optional
        Generally, each XML file contains data for a single borehole but in some cases
        multiple boreholes can be present in the XML file. If set to True, all boreholes
        will be read from the XML file. The default is False, which reads only the first.
        A warning will be raised if multiple boreholes are found in the XML file and
        read_all is False.

    Returns
    -------
    dict | list[dict]
        Dict or list of dictionaries containing the extracted data from the XML file. If
        the XML file contains data for a single borehole, a dictionary will be returned.
        If multiple boreholes are found and `read_all` is True, a list of dictionaries
        will be returned.

    Raises
    ------
    ValueError
        If no predefined schema is found for the specified company.
    SyntaxError
        If the XML file does not conform to the expected schema.

    """
    company = company or "BRO"

    if schema is None:
        try:
            schema = schemas.bhrg[company]
        except KeyError as e:
            raise ValueError(
                f"No predefined schema for '{company}' in BHR-G. Supported companies are: "
                f"{schemas.bhrg.keys()}. Define a custom schema or use a supported company."
            ) from e

    try:
        result = _read_xml(file, schema, schema.get("payload_root", None), read_all)
    except SyntaxError as e:
        raise SyntaxError(f"Invalid xml schema for XML file: {file}.") from e

    return result


def read_sfr(
    file: bytes | str | Path,
    company: str = None,
    schema: dict[str, Any] = None,
    read_all: bool = False,
) -> dict | list[dict]:
    """
    Read Soil outcrop descriptions (SFR) from an XML file or bytestring and extract
    relevant data based on a predefined or custom schema that describes the XML structure.

    Parameters
    ----------
    file : bytes | str | Path
        XML filepath-like or bytestring containing the XML data to read.
    company : str, optional
        Specify a company name to use a predefined schema for that company if available.
        See `schemas.sfr` for available companies. The default is None, then company will
        default to "BRO" and the predefined BRO schema will be used if no custom schema
        is defined.
    schema : dict[str, Any], optional
        Custom schema used to parse the XML structure.
    read_all : bool, optional
        Generally, each XML file contains data for a single outcrop but in some cases
        multiple outcrops can be present in the XML file. If set to True, all outcrops
        will be read from the XML file. The default is False, which reads only the first.
        A warning will be raised if multiple outcrops are found in the XML file and
        read_all is False.

    Returns
    -------
    dict | list[dict]
        Dict or list of dictionaries containing the extracted data from the XML file. If
        the XML file contains data for a single outcrop, a dictionary will be returned.
        If multiple outcrops are found and `read_all` is True, a list of dictionaries
        will be returned.

    Raises
    ------
    ValueError
        If no predefined schema is found for the specified company.
    SyntaxError
        If the XML file does not conform to the expected schema.

    """
    company = company or "BRO"

    if schema is None:
        try:
            schema = schemas.sfr[company]
        except KeyError as e:
            raise ValueError(
                f"No predefined schema for '{company}' in SFR. Supported companies are: "
                f"{schemas.sfr.keys()}. Define a custom schema or use a supported company."
            ) from e

    try:
        result = _read_xml(file, schema, schema.get("payload_root", None), read_all)
    except SyntaxError as e:
        raise SyntaxError(f"Invalid xml schema for XML file: {file}.") from e

    return result



from pathlib import Path

import geopandas as gpd
import pandas as pd

from geost.utils import create_connection


class Geopackage:
    """
    Read and inspect the contents of a geopackage file. The class is designed to be used
    as a context manager to ensure the connection to the geopackage is closed after use.

    Parameters
    ----------
    file : str | Path
        Valid string path to the Geopackage file.

    Examples
    --------
    Check the available layers in a geopackage file:

    >>> print(Geopackage("my_geopackage.gpkg").layers())

    Read the first five records of a table in the geopackage for a quick inspection:

    >>> with Geopackage("my_geopackage.gpkg") as gp:
    ...     print(gp.table_head("my_table"))

    For use without a context manager:

    >>> gp = Geopackage("my_geopackage.gpkg")
    >>> gp.get_connection()
    >>> table = gp.read_table("my_table") # Returns a DataFrame of the table.

    """

    def __init__(self, file: str | Path):
        self.file = file
        self.connection = None

    def __enter__(self):
        self.get_connection()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def _get_cursor(self):
        """
        Get a sqlite3 cursor object for the geopackage connection to execute database
        queries with. This method is used internally to execute queries and should not
        be used directly.

        """
        return self.connection.cursor()

    def layers(self) -> pd.DataFrame:
        """
        Return a Pandas DataFrame containing the layers and associated geometry types in
        the geopackage.

        """
        return gpd.list_layers(self.file)

    def get_connection(self):
        """
        Create a manual sqlite3 connection to the geopackage file. Establishing and closing
        this connection is automatically handled when the class is used as a context manager.

        """
        self.connection = create_connection(self.file)

    def close(self):
        """
        Close the connection to the geopackage file.

        """
        if self.connection:
            self.connection = None

    def get_column_names(self, table: str) -> list:
        """
        Get the column names of a table in the geopackage.

        Parameters
        ----------
        table : string
            Name of the table to get the column names for.

        Returns
        -------
        columns : list
            List of the column names for the table.

        """
        cursor = self._get_cursor()
        cursor.execute(f"SELECT * FROM {table}")
        columns = [col[0] for col in cursor.description]
        return columns

    def table_head(self, table: str) -> pd.DataFrame:
        """
        Select the first five records from a table the in geopackage for quick inspection.

        Parameters
        ----------
        table : string
            Name of the table to select the first records from.

        Returns
        -------
        pd.DataFrame
            Pandas DataFrame of the first five records.

        """
        cursor = self._get_cursor()
        cursor.execute(f"SELECT * FROM {table} LIMIT 5")
        data = cursor.fetchall()
        return pd.DataFrame(data, columns=self.get_column_names(table))

    def read_table(self, table: str) -> pd.DataFrame:
        """
        Read all data in a specified table of the geopackage.

        Parameters
        ----------
        table : str
            Name of the table to read. Check the available tables with `Geopackage.layers()`.

        Returns
        -------
        pd.DataFrame
            Pandas DataFrame of the table.

        """
        columns = self.get_column_names(table)
        table = self.query(f"SELECT * FROM {table}", columns)
        return table

    def query(self, query: str, outcolumns: list = None) -> pd.DataFrame:
        """
        Use a custom query on the geopackage to retrieve desired tables or joined information
        from multiple tables.

        Parameters
        ----------
        query : str
            Full string of the SQL query to retrieve the desired table with.
        outcolumns : list, optional
            Specify column names to be used for the output table. The default is
            None.

        Returns
        -------
        pd.DataFrame
            Result DataFrame of the query.

        """
        cursor = self._get_cursor()
        cursor.execute(query)
        data = cursor.fetchall()

        if outcolumns is None:
            outcolumns = [col[0] for col in cursor.description]

        return pd.DataFrame(data, columns=outcolumns)



import numpy as np
import xarray as xr


def get_bbox_criteria(xmin, ymin, xmax, ymax):
    json_line = {
        "area": {
            "boundingBox": {
                "lowerCorner": {"lat": ymin, "lon": xmin},
                "upperCorner": {"lat": ymax, "lon": xmax},
            }
        }
    }
    return json_line


def coordinates_to_cellcenters(
    ds: xr.Dataset | xr.DataArray, cellsize: int | float, dz: int | float = None
) -> xr.Dataset | xr.DataArray:
    """
    Change the coordinates of an Xarray Dataset or DataArray to cellcenters. This assumes
    that each coordinate in the Dataset are lower left, lower z (if "z" dimension is present)
    coordinates and y-coordinates are in increasing order which is the case in BRO distributed
    models.

    Parameters
    ----------
    ds : xr.Dataset | xr.DataArray
        Dataset or DataArray to change the coordinates for.
    cellsize : int | float
        Cellsize of the Dataset or DataArray.
    dz : int | float, optional
        Size of the "z" dimension if present in the Dataset or DataArray. If a "z" dimension
        is present but dz is None, the dimension is ignored. The default is None.

    Returns
    -------
    xr.Dataset | xr.DataArray
        Dataset or DataArray with the changed coordinates.

    """
    ds["x"] = ds["x"] + (cellsize / 2)
    ds["y"] = ds["y"] + (cellsize / 2)
    if dz:
        ds["z"] = ds["z"] + (dz / 2)
    return ds


def flip_ycoordinates(ds):
    if "z" in ds.dims:
        ds = ds.transpose("y", "x", "z")
    else:
        ds = ds.transpose("y", "x")

    if ds["y"][-1] > ds["y"][0]:
        ds = ds.sel(y=slice(None, None, -1))

    return ds



from itertools import product
from pathlib import Path

import geopandas as gpd
import numpy as np
import pandas as pd
import pytest
import xarray as xr
from shapely import geometry as gmt

from geost import read_borehole_table, read_nlog_cores
from geost.base import DiscreteData, LayeredData
from geost.models.basemodels import VoxelModel


def borehole_a():
    nlayers = 5
    top = [0, 0.8, 1.5, 2.5, 3.7]
    bottom = top[1:] + [4.2]
    mv = 0.2
    end = mv - bottom[-1]
    return pd.DataFrame(
        {
            "nr": np.full(nlayers, "A"),
            "x": np.full(nlayers, 2),
            "y": np.full(nlayers, 3),
            "surface": np.full(nlayers, mv),
            "end": np.full(nlayers, end),
            "top": top,
            "bottom": bottom,
            "lith": ["K", "K", "Z", "Z", "K"],
        }
    )


def borehole_b():
    nlayers = 5
    top = [0, 0.6, 1.2, 2.5, 3.1]
    bottom = top[1:] + [3.9]
    mv = 0.3
    end = mv - bottom[-1]
    return pd.DataFrame(
        {
            "nr": np.full(nlayers, "B"),
            "x": np.full(nlayers, 1),
            "y": np.full(nlayers, 4),
            "surface": np.full(nlayers, mv),
            "end": np.full(nlayers, end),
            "top": top,
            "bottom": bottom,
            "lith": ["K", "K", "V", "V", "K"],
        }
    )


def borehole_c():
    nlayers = 5
    top = [0, 1.4, 1.8, 2.9, 3.8]
    bottom = top[1:] + [5.5]
    mv = 0.25
    end = mv - bottom[-1]
    return pd.DataFrame(
        {
            "nr": np.full(nlayers, "C"),
            "x": np.full(nlayers, 4),
            "y": np.full(nlayers, 2),
            "surface": np.full(nlayers, mv),
            "end": np.full(nlayers, end),
            "top": top,
            "bottom": bottom,
            "lith": ["K", "K", "K", "Z", "Z"],
        }
    )


def borehole_d():
    nlayers = 5
    top = [0, 0.5, 1.2, 1.8, 2.5]
    bottom = top[1:] + [3.0]
    mv = 0.1
    end = mv - bottom[-1]
    return pd.DataFrame(
        {
            "nr": np.full(nlayers, "D"),
            "x": np.full(nlayers, 3),
            "y": np.full(nlayers, 5),
            "surface": np.full(nlayers, mv),
            "end": np.full(nlayers, end),
            "top": top,
            "bottom": bottom,
            "lith": ["K", "V", "K", "V", "Z"],
        }
    )


def borehole_e():
    nlayers = 5
    top = [0, 0.5, 1.2, 1.8, 2.5]
    bottom = top[1:] + [3.0]
    mv = -0.1
    end = mv - bottom[-1]
    return pd.DataFrame(
        {
            "nr": np.full(nlayers, "E"),
            "x": np.full(nlayers, 1),
            "y": np.full(nlayers, 1),
            "surface": np.full(nlayers, mv),
            "end": np.full(nlayers, end),
            "top": top,
            "bottom": bottom,
            "lith": ["Z", "Z", "Z", "Z", "Z"],
        }
    )


@pytest.fixture
def testdatadir():
    return Path(__file__).parent / "data"


@pytest.fixture
def borehole_file(testdatadir):
    return testdatadir / r"test_boreholes.parquet"


@pytest.fixture
def nlog_borehole_file(testdatadir):
    return testdatadir / r"test_nlog_stratstelsel_20230807.parquet"


@pytest.fixture
def borehole_collection(borehole_data):
    """
    Fixture containing a BoreholeCollection instance of synthetic borehole data.

    """
    borehole_collection = borehole_data.to_collection()
    return borehole_collection


@pytest.fixture
def nlog_borehole_collection(nlog_borehole_file):
    nlog_borehole_collection = read_nlog_cores(nlog_borehole_file)
    return nlog_borehole_collection


@pytest.fixture
def borehole_data():
    a = borehole_a()
    b = borehole_b()
    c = borehole_c()
    d = borehole_d()
    e = borehole_e()
    df = pd.concat([a, b, c, d, e], ignore_index=True)
    """
    Fixture containing a LayeredData instance of synthetic borehole data.

    """
    return LayeredData(df)


# Fixtures for header testing
@pytest.fixture
def point_header_gdf():
    """
    Creates a synthetic header geodataframe for testing
    """
    x_coors = [1.0, 2.0, 3.0, 4.0, 5.0]
    y_coors = [1.0, 2.0, 3.0, 4.0, 5.0]
    coordinates = np.array([(x, y) for x, y in product(x_coors, y_coors)])
    nrs = ["nr" + str(i + 1) for i in range(len(coordinates))]
    mvs = np.arange(1, 26)
    ends = np.arange(-1, -26, -1)
    geometries = [gmt.Point(c) for c in coordinates]
    gdf = gpd.GeoDataFrame(
        {
            "nr": nrs,
            "x": coordinates[:, 0],
            "y": coordinates[:, 1],
            "surface": mvs,
            "end": ends,
        },
        geometry=geometries,
    )
    gdf.set_crs("epsg:28992", inplace=True)
    return gdf


def cpt_a():
    """
    Helper function for a synthetic CPT containing qs, fs and u2 "measurements".

    """
    depth = np.arange(10)
    surface = 2.1
    end = surface - depth.max()
    qc = [0.227, 0.279, 0.327, 0.354, 0.357, 0.354, 0.363, 0.447, 0.761, 1.481]
    fs = [0.010, 0.014, 0.019, 0.021, 0.022, 0.023, 0.026, 0.023, 0.022, 0.021]
    u2 = [0.018, 0.026, 0.035, 0.041, 0.047, 0.052, 0.058, 0.061, 0.057, 0.036]
    return pd.DataFrame(
        {
            "nr": np.repeat("a", 10),
            "x": np.repeat(1, 10),
            "y": np.repeat(1, 10),
            "surface": np.repeat(surface, 10),
            "end": np.repeat(end, 10),
            "depth": depth,
            "qc": qc,
            "fs": fs,
            "u2": u2,
        }
    )


def cpt_b():
    """
    Helper function for a synthetic CPT containing qs, fs and u2 "measurements".

    """
    depth = np.arange(10)
    surface = 0.8
    end = surface - depth.max()
    qc = [8.721, 12.733, 17.324, 17.036, 16.352, 15.781, 15.365, 15.509, 15.884, 15.982]
    fs = [0.061, 0.058, 0.055, 0.054, 0.052, 0.051, 0.052, 0.051, 0.051, 0.050]
    u2 = [0.218, 0.219, 0.221, 0.220, 0.219, 0.220, 0.221, 0.224, 0.224, 0.225]
    return pd.DataFrame(
        {
            "nr": np.repeat("b", 10),
            "x": np.repeat(2, 10),
            "y": np.repeat(2, 10),
            "surface": np.repeat(surface, 10),
            "end": np.repeat(end, 10),
            "depth": depth,
            "qc": qc,
            "fs": fs,
            "u2": u2,
        }
    )


@pytest.fixture
def cpt_data():
    """
    Fixture containing a DiscreteData instance of synthetic CPT data.

    """
    df = pd.concat([cpt_a(), cpt_b()], ignore_index=True)
    return DiscreteData(df)


@pytest.fixture
def cpt_collection(cpt_data):
    return cpt_data.to_collection()


@pytest.fixture
def xarray_dataset():
    x = np.arange(4) + 0.5
    y = x[::-1]
    z = np.arange(-2, 0, 0.5) + 0.25

    strat = [
        [[2, 2, 2, 1], [2, 2, 1, 1], [2, 1, 1, 1], [2, 2, 1, 1]],
        [[2, 2, 1, 1], [2, 2, 1, 1], [2, 1, 1, 1], [2, 1, 2, 1]],
        [[2, 2, 2, 1], [2, 1, 1, 1], [2, 2, 1, 1], [2, 2, 2, 1]],
        [[2, 2, 2, 1], [2, 1, 1, 1], [2, 2, 1, 1], [2, 2, 2, 1]],
    ]
    lith = [
        [[2, 3, 2, 1], [2, 3, 1, 1], [2, 1, 1, 1], [3, 2, 1, 1]],
        [[2, 2, 1, 1], [2, 2, 1, 1], [2, 1, 1, 1], [2, 3, 2, 1]],
        [[2, 3, 2, 1], [2, 1, 1, 3], [2, 2, 1, 1], [2, 2, 2, 1]],
        [[2, 2, 2, 1], [2, 1, 1, 1], [2, 2, 1, 3], [3, 2, 2, 1]],
    ]
    ds = xr.Dataset(
        data_vars=dict(strat=(["y", "x", "z"], strat), lith=(["y", "x", "z"], lith)),
        coords=dict(y=y, x=x, z=z),
    )
    ds.rio.write_crs(28992, inplace=True)
    return ds


@pytest.fixture
def voxelmodel(xarray_dataset):
    return VoxelModel(xarray_dataset)


def create_polygons() -> gpd.GeoDataFrame:
    """
    Helper function to create a GeoDataFrame with 10 random irregular shaped polygons to
    create test fixtures with.

    Returns
    -------
    gpd.GeoDataFrame
        GeoDataFrame with only a "geometry" column containing the polygons.

    """
    points = [
        [0, 0],
        [0, 4],
        [4, 4],
        [4, 0],
        [0.5, 0.5],
        [0.7, 3],
        [1, 1.5],
        [2.1, 2.3],
        [2.7, 0.8],
        [3.1, 3.1],
        [3.3, 1.9],
        [1.8, 3.5],
        [1.5, 0.8],
        [1.4, 1.0],
    ]
    points = gpd.GeoDataFrame(geometry=[gmt.Point(p) for p in points], crs=28992)
    points["geometry"] = points.voronoi_polygons()
    points = points.clip((0, 0, 4, 4), sort=True)
    return points


@pytest.fixture
def simple_soilmap_gpkg(tmp_path):
    """
    Fixture to create a tmp geopackage file that contains relevant BRO soilmap information
    to test.

    """
    polygons = create_polygons()
    maparea_id = np.arange(len(polygons))
    soilunits = [
        "pVc",  # Peat type
        "hVk",  # Peat type
        "kVc",  # Peat type
        "Vc",  # Peat type
        "AAP",  # Peat type
        "vWp",  # Moerig type
        "iWp",  # Moerig type
        "kWz",  # Moerig type
        "AWv",  # Moerig type
        "Rv01C",  # Buried type
        "pRv81",  # Buried type
        "Mv51A",  # Buried type
        "Mv81A",  # Buried type
        "bEZ23",  # Other type
    ]
    polygons["maparea_id"] = maparea_id
    soilcodes = gpd.GeoDataFrame({"maparea_id": maparea_id, "soilunit_code": soilunits})

    layers = ["soilarea", "soilarea_soilunit"]
    tables = [polygons, soilcodes]

    outfile = tmp_path / "soilmap.gpkg"
    for layer, table in zip(layers, tables):
        table.to_file(outfile, driver="GPKG", layer=layer, index=False)

    return outfile


@pytest.fixture
def bro_cpt_gpkg(testdatadir):
    """
    Small extraction of 4 CPTs from the BRO CPT geopackage for testing purposes. The CPTs
    were selected from the original BRO CPT geopackage by their primary keys. The selected
    keys were: 164970, 164971, 164975, 164976. These numbers coincide with the "fid" index
    in the GeoDataFrame when `geost.bro.BroCptGeopackage` is used to read the geopackage.

    """
    return testdatadir / r"test_bro_cpt_geopackage.gpkg"



from geost import data
from geost.io.read import (
    bro_api_read,
    read_bhrg,
    read_bhrgt,
    read_bhrp,
    read_borehole_table,
    read_collection_geopackage,
    read_cpt,
    read_cpt_table,
    read_gef_cpts,
    read_nlog_cores,
    read_pickle,
    read_sfr,
    read_uullg_tables,
    read_xml_boris,
)
from geost.utils import csv_to_parquet, excel_to_parquet

__version__ = "0.3.0"



import warnings

import pandas as pd
from pandera.errors import SchemaError, SchemaErrors
from pandera.pandas import DataFrameSchema

from geost import config
from geost._warnings import ValidationWarning


def safe_validate(schema: DataFrameSchema, df: pd.DataFrame, **kwargs) -> pd.DataFrame:
    """
    Catch validation errors and raise a warning instead while returning the DataFrame
    unchanged. If rows are dropped due to validation, report which rows were dropped.

    Parameters
    ----------
    schema : DataFrameSchema
        The schema to validate the DataFrame against.
    df : pd.DataFrame
        The DataFrame to validate.
    **kwargs : keyword arguments
        Additional arguments to pass to the Pandera DataFrameSchema.validate method. Note
        that the inplace argument is ignored and the DataFrame will not be modified in place.
        i.e. this function always returns a DataFrame.

    Returns
    -------
    pd.DataFrame
        Either the validated (and possibly changed) DataFrame or the unchanged DataFrame
        if validation failed.
    """
    if "inplace" in kwargs:
        raise ValueError("'inplace' argument is not supported.")

    # Update schema drop invalid rows setting, adjust lazy validation accordingly
    # (Pandera requires lazy validation when drop_invalid_rows is True)
    schema.drop_invalid_rows = config.validation.DROP_INVALID
    kwargs["lazy"] = True if config.validation.DROP_INVALID else False

    try:
        validated_df = schema.validate(df, **kwargs)
        if config.validation.DROP_INVALID:
            dropped = df.index.difference(validated_df.index)
            if len(dropped) > 0 and config.validation.VERBOSE:
                warnings.warn(
                    f"\nValidation dropped {len(dropped)} row(s) for schema '{schema.name}'.\n"
                    f"Dropped indices: {list(dropped)}\n",
                    category=ValidationWarning,
                )
        return validated_df
    except (SchemaError, SchemaErrors) as e:
        if config.validation.VERBOSE:
            warnings.warn(
                f"\nValidation failed for schema '{schema.name}'.\nDetails:\n{str(e)}\n",
                category=ValidationWarning,
            )
        if config.validation.FLAG_INVALID:
            df["is_valid"] = ~df.index.isin(e.failure_cases["index"])
        return df



import numpy as np
import xarray as xr
from shapely.geometry import LineString


def sample_with_coords(ds: xr.Dataset | xr.DataArray, coords: np.ndarray):
    """
    Sample x and y dims of an Xarray Dataset or DataArray using an array of x and y
    coordinates. Only coordinates that are within the extent of the Dataset or DataArray
    are selected.

    Parameters
    ----------
    ds : xr.Dataset or xr.DataArray
        Dataset or DataArray to sample. Must contain dimensions 'x' and 'y' that
        refer to the coordinates.
    coords : np.ndarray
        Numpy array with shape (n, 2) of the x and y coordinates to select.

    Returns
    -------
    ds_sel : xr.Dataset or xr.DataArray
        Sampled Dataset or DataArray with dimension 'idx' which is the index of the input
        coordinates which are selected.

    """
    x, y = coords[:, 0], coords[:, 1]
    idx = np.arange(len(coords))

    xmin, ymin, xmax, ymax = ds.rio.bounds()
    outside_bbox = (x < xmin) | (x > xmax) | (y < ymin) | (y > ymax)

    idx = idx[~outside_bbox]
    x = x[~outside_bbox]
    y = y[~outside_bbox]

    ds_sel = ds.sel(
        x=xr.DataArray(x, dims="idx"), y=xr.DataArray(y, dims="idx"), method="nearest"
    )
    ds_sel = ds_sel.assign_coords(idx=("idx", idx))
    return ds_sel


def sample_along_line(
    ds: xr.Dataset | xr.DataArray,
    line: LineString,
    dist: int | float = None,
    nsamples: int = None,
):
    """
    Sample x and y dims of an Xarray Dataset or DataArray over distance along a
    Shapely LineString object. Sampling can be done using a specified distance
    or a specified number of samples that need to be taken along the line.

    Parameters
    ----------
    ds : xr.Dataset or xr.DataArray
        Dataset or DataArray to sample. Must contain dimensions 'x' and 'y' that
        refer to the coordinates.
    line : LineString
        shapely.geometry.LineString object to use for sampling.
    dist : int, float, optional
        Distance between each sample along the line. Takes equally distant samples
        from the start of the line untill it reaches the end. The default is None.
    nsamples : int, optional
        Number of samples to take along the line between the beginning and the end.
        The default is None.

    Raises
    ------
    ValueError
        If both or none of dist and nsamples are specified.

    Returns
    -------
    ds_sel : xr.Dataset or xr.DataArray
        Sampled Dataset or DataArray with dimension 'dist' for distance.

    """
    if dist and nsamples:
        raise ValueError("Cannot use 'dist' and 'nsamples' together, use one option.")

    elif dist:
        sample_locs = np.arange(0, line.length, dist)

    elif nsamples:
        sample_locs = np.linspace(0, line.length, nsamples)

    else:
        raise ValueError("'dist' or 'nsamples' not specified, use one option.")

    samplepoints = np.array([_interpolate_point(line, loc) for loc in sample_locs])
    dist, x, y = samplepoints[:, 0], samplepoints[:, 1], samplepoints[:, 2]

    ds_sel = ds.sel(
        x=xr.DataArray(x, dims="dist"), y=xr.DataArray(y, dims="dist"), method="nearest"
    )
    ds_sel = ds_sel.assign_coords(dist=("dist", dist))

    return ds_sel


def _interpolate_point(line, loc):
    """
    Return the location (i.e. distance) and x and y coordinates of an interpolated
    point along a Shapely LineString object.

    Parameters
    ----------
    line : LineString
        shapely.geometry.LineString object.
    loc : int, float
        Distance along the LineString to interpolate the point at.

    """
    p = line.interpolate(loc)
    return loc, p.x, p.y


def label_consecutive_2d(array: np.ndarray, axis: int = 0) -> np.ndarray:
    """
    Label consecutive array elements with unique numbers along a specified axis in a 2D
    array.

    Parameters
    ----------
    array : np.ndarray
        Array of shape (m, n) to label the elements in.
    axis : int, optional
        Axis along to label. The default is 0.

    Returns
    -------
    np.ndarray
        Array with labelled elements with the same shape as the input array.

    """
    labels = np.cumsum(np.diff(array, axis=axis) != 0, axis=axis)

    if axis == 0:
        start_labels = np.zeros(array.shape[1], dtype=array.dtype).reshape(1, -1)
        labels = np.r_[start_labels, labels]
    else:
        start_labels = np.zeros(array.shape[0], dtype=array.dtype).reshape(-1, 1)
        labels = np.c_[start_labels, labels]

    return labels



import re
from collections import defaultdict
from contextlib import suppress
from typing import Any

import numpy as np
from lxml import etree


def parse_crs(el: etree.Element) -> str:
    """
    Retrieve the coordinate reference system (CRS) from an XML element.

    Parameters
    ----------
    el : etree.Element
        The XML element containing the CRS information.

    Returns
    -------
    str
        String representing the CRS (e.g. 'urn:ogc:def:crs:EPSG::28992'), or "unknown"
        if not found.

    """
    return el.attrib.get("srsName", "unknown")


def parse_coordinates(coords: str) -> tuple[float, float]:
    """
    Parse a string containing coordinates as a tuple of floats.

    Parameters
    ----------
    coords : str
        Any of {'x y', 'x,y', 'x;y'} where x and y are parsable by float

    Returns
    -------
    tuple[float, float]
        The parsed coordinates as a tuple of floats.

    """
    if " " in coords:
        splitter = r"\s+"
    elif "," in coords:
        splitter = ","
    elif ";" in coords:
        splitter = ";"
    else:
        raise ValueError(f"Cannot parse coordinates: '{coords}'")

    try:
        x, y = re.split(splitter, coords)
    except ValueError as e:
        raise ValueError(f"Cannot parse coordinates: '{coords}'") from e

    return float(x), float(y)


def safe_float(value: Any) -> float | None:
    """
    Try to cast a value to a float dtype. Returns None if the value cannot be casted.

    Parameters
    ----------
    value : Any
        The value to cast.

    Returns
    -------
    float | None
        Value as a float type or None if the value cannot be casted into a float.

    """
    with suppress(TypeError, ValueError):
        return float(value)


def clean_string(val: str) -> str:
    """
    Clean a string to keep only letters, numbers and punctuation.

    Parameters
    ----------
    val : str
        The input string to clean.

    Returns
    -------
    str
        The cleaned string with unwanted characters removed.

    """
    return re.sub(r"[^\w.,:;!?()-]", "", val)


def safe_get(el: etree.Element):
    """
    Safely get the text from an `etree.Element` instance. Returns `None` if it would raise
    an `AttributeError`. This occurs for example when `el` is a NoneType.

    """
    with suppress(AttributeError):
        value = el.text
        if value is None:
            return None
        else:
            return clean_string(value)


def _process_layers(layers: list[etree.Element], attributes: list[str]) -> dict:
    """
    Process a list of XML elements representing layers and extract specified attributes.

    Parameters
    ----------
    layers : list[etree.Element]
        List of XML elements representing layers.
    attributes : list[str]
        List of attribute names to extract from each layer.

    Returns
    -------
    dict
        Dictionary with the searched layer attributes as keys and lists of each value per
        layer.

    """
    data = defaultdict(list)
    for layer in layers:
        for attr in attributes:
            attribute = layer.xpath(f".//*[local-name() = '{attr}']")
            value = safe_get(attribute[0]) if attribute else None
            data[attr].append(value)

    return data


def process_bhrgt_data(el: etree.Element, attributes: list | None) -> dict:
    """
    Process an XML element containing the layer descriptions in BHR-GT data objects.

    Parameters
    ----------
    el : etree.Element
        Element containing the layer descriptions.
    attributes : list[str] | None
        List with string names of the attributes to retrieve from each layer. If the input
        is None, it will be attempted to at least retrieve "upperBoundary", "lowerBoundary"
        and "geotechnicalSoilName" from each layer.

    Returns
    -------
    dict
        Dictionary with the searched layer attributes as keys and lists of each value per
        layer.

    """
    if attributes is None:
        attributes = ["upperBoundary", "lowerBoundary", "geotechnicalSoilName"]

    layers = el.xpath(".//*[local-name() = 'layer']")

    return _process_layers(layers, attributes)


def process_bhrg_data(el: etree.Element, attributes: list | None) -> dict:
    """
    Process an XML element containing the layer descriptions in BHR-G data objects.

    Parameters
    ----------
    el : etree.Element
        Element containing the layer descriptions.
    attributes : list[str] | None
        List with string names of the attributes to retrieve from each layer. If the input
        is None, it will be attempted to at least retrieve "upperBoundary", "lowerBoundary"
        and "soilNameNEN5104" from each layer.

    Returns
    -------
    dict
        Dictionary with the searched layer attributes as keys and lists of each value per
        layer.

    """
    if attributes is None:
        attributes = ["upperBoundary", "lowerBoundary", "soilNameNEN5104"]

    layers = el.xpath(".//*[local-name() = 'layer']")

    return _process_layers(layers, attributes)


def process_bhrp_data(el: etree.Element, attributes: list | None) -> dict:
    """
    Process an XML element containing the layer descriptions in BHR-P data objects.

    Parameters
    ----------
    el : etree.Element
        Element containing the layer descriptions.
    attributes : list[str] | None
        List with string names of the attributes to retrieve from each layer. If the input
        is None, it will be attempted to at least retrieve "upperBoundary", "lowerBoundary"
        and "standardSoilName" from each layer.

    Returns
    -------
    dict
        Dictionary with the searched layer attributes as keys and lists of each value per
        layer.

    """
    if attributes is None:
        attributes = ["upperBoundary", "lowerBoundary", "standardSoilName"]

    layers = el.xpath(".//*[local-name() = 'soilLayer']")

    return _process_layers(layers, attributes)


def process_cpt_data(el: etree.Element) -> dict:
    """
    Process an XML element containing the data in CPT data objects.

    Parameters
    ----------
    el : etree.Element
        Element containing the CPT data.

    Returns
    -------
    dict
        Dictionary with the processed CPT data.

    """
    cpt_data = el.xpath(".//*[local-name() = 'cptResult']")[0]
    parameters = el.xpath(".//*[local-name() = 'parameters']")[0]

    encoding = cpt_data.xpath(".//*[local-name() = 'TextEncoding']")[0]
    row_sep = encoding.attrib.get("blockSeparator", ",")
    column_sep = encoding.attrib.get("tokenSeparator", ";")

    values = cpt_data.xpath(".//*[local-name() = 'values']")[0]
    values = textvalues_to_array(values.text, column_sep, row_sep)

    missing_value = -999999
    values[values == missing_value] = np.nan

    data = {}
    for ii, param in enumerate(parameters.iterchildren()):
        data[strip_tag(param)] = values[:, ii]

    return data


def process_sfr_data(el: etree.Element, attributes: list | None) -> dict:
    """
    Process an XML element containing the layer descriptions in SFR data objects.

    Parameters
    ----------
    el : etree.Element
        Element containing the layer descriptions.
    attributes : list[str] | None
        List with string names of the attributes to retrieve from each layer. If the input
        is None, it will be attempted to at least retrieve "upperBoundary", "lowerBoundary"
        and "soilNameNEN5104" from each layer.

    Returns
    -------
    dict
        Dictionary with the searched layer attributes as keys and lists of each value per
        layer.

    """
    if attributes is None:
        attributes = ["upperBoundary", "lowerBoundary", "soilNameNEN5104"]

    layers = el.xpath(".//*[local-name() = 'SoilLayer']")

    return _process_layers(layers, attributes)


def strip_tag(el: etree.Element) -> str:
    """
    Strip namespaces from the tag name of an XML element, returning only the local name.

    Parameters
    ----------
    el : etree.Element
        The XML element to strip.

    Returns
    -------
    str
        The local name of the element without any namespace prefix.

    """
    tag = el.tag.split("}")[-1] if "}" in el.tag else el.tag
    return tag.lower()


def textvalues_to_array(text: str, column_sep: str, row_sep: str) -> list:
    """
    Convert a string of text values into a list of lists, splitting by column and row
    separators.

    Parameters
    ----------
    text : str
        The input string containing text values.
    column_sep : str, optional
        The separator for columns (default is ',').
    row_sep : str, optional
        The separator for rows (default is ';').

    Returns
    -------
    list
        A list of lists containing the split text values.

    """
    ncols = text.split(row_sep)[0].count(column_sep) + 1

    if column_sep == ".":  # np.fromstring does not support '.' as a separator
        text = text.replace(column_sep, row_sep)
        sep = row_sep
    else:
        text = text.replace(row_sep, column_sep)
        sep = column_sep

    array = np.fromstring(text, sep=sep, dtype=float)

    return array.reshape(-1, ncols)



from pathlib import Path

import numpy as np
import pandas as pd
import pytest
from numpy.testing import assert_array_almost_equal, assert_array_equal
from pyvista import MultiBlock, UnstructuredGrid
from shapely import get_coordinates

from geost.base import (
    BoreholeCollection,
    CptCollection,
    DiscreteData,
    LayeredData,
    PointHeader,
)
from geost.data_objects import Cpt
from geost.export import geodataclass


class TestLayeredData:
    @pytest.mark.unittest
    def test_datatype(self, borehole_data):
        assert borehole_data.datatype == "layered"

    @pytest.mark.unittest
    def test_to_header(self, borehole_data):
        expected_columns = ["nr", "x", "y", "surface", "end", "geometry"]

        header = borehole_data.to_header()

        assert isinstance(header, PointHeader)
        assert_array_equal(header.gdf.columns, expected_columns)
        assert len(header.gdf) == 5
        assert header["nr"].nunique() == 5
        assert header.horizontal_reference == 28992
        assert header.vertical_reference == 5709

    @pytest.mark.unittest
    def test_to_collection(self, borehole_data):
        collection = borehole_data.to_collection()
        assert isinstance(collection, BoreholeCollection)
        assert isinstance(collection.header, PointHeader)
        assert len(collection.header) == 5

    @pytest.mark.unittest
    def test_select_by_values(self, borehole_data):
        selected = borehole_data.select_by_values("lith", ["V", "K"], how="or")
        assert isinstance(selected, LayeredData)

        expected_nrs = ["A", "B", "C", "D"]
        selected_nrs = selected["nr"].unique()
        assert_array_equal(selected_nrs, expected_nrs)

        selected = borehole_data.select_by_values("lith", ["V", "K"], how="and")

        expected_nrs = ["B", "D"]
        selected_nrs = selected["nr"].unique()

        assert_array_equal(selected_nrs, expected_nrs)

    @pytest.mark.unittest
    def test_slice_by_values(self, borehole_data):
        sliced = borehole_data.slice_by_values("lith", "Z")
        assert isinstance(sliced, LayeredData)

        expected_boreholes_with_sand = ["A", "C", "D", "E"]
        expected_length = 10

        assert_array_equal(sliced["nr"].unique(), expected_boreholes_with_sand)
        assert np.all(sliced["lith"] == "Z")
        assert len(sliced) == expected_length

        sliced = borehole_data.slice_by_values("lith", "Z", invert=True)

        expected_boreholes_without_sand = ["A", "B", "C", "D"]
        expected_length = 15

        assert_array_equal(sliced["nr"].unique(), expected_boreholes_without_sand)
        assert ~np.any(sliced["lith"] == "Z")
        assert len(sliced) == expected_length

    @pytest.mark.unittest
    def test_get_cumulative_thickness(self, borehole_data):
        result = borehole_data.get_cumulative_thickness("lith", "V")
        expected_boreholes_returned = ["B", "D"]
        expected_thickness = [1.9, 1.4]

        assert len(result) == 2
        assert_array_equal(result.index, expected_boreholes_returned)
        assert_array_almost_equal(result["V"], expected_thickness)

        result = borehole_data.get_cumulative_thickness("lith", ["Z", "K"])
        expected_boreholes_returned = ["A", "B", "C", "D", "E"]
        expected_sand_thickness = [2.2, np.nan, 2.6, 0.5, 3.0]
        expected_clay_thickness = [2.0, 2.0, 2.9, 1.1, np.nan]

        assert result.shape == (5, 2)
        assert_array_equal(result.index, expected_boreholes_returned)
        assert_array_almost_equal(result["K"], expected_clay_thickness)
        assert_array_almost_equal(result["Z"], expected_sand_thickness)

    @pytest.mark.unittest
    def test_get_layer_top(self, borehole_data):
        result = borehole_data.get_layer_top("lith", "V")
        expected_boreholes_returned = ["B", "D"]
        expected_tops = [1.2, 0.5]

        assert len(result) == 2
        assert_array_equal(result.index, expected_boreholes_returned)
        assert_array_almost_equal(result["V"], expected_tops)

        # Test with a minimum depth to start the search for the top of the layer.
        result = borehole_data.get_layer_top("lith", "V", min_depth=2)
        expected_boreholes_returned = ["B", "D"]
        expected_tops = [2.0, 2.0]

        assert len(result) == 2
        assert_array_equal(result.index, expected_boreholes_returned)
        assert_array_almost_equal(result["V"], expected_tops)

        # Test with a minimum depth and minimum thickness
        result = borehole_data.get_layer_top(
            "lith", "V", min_depth=2, min_thickness=0.6
        )
        expected_boreholes_returned = ["B"]
        expected_tops = [2.5]

        assert len(result) == 1
        assert_array_equal(result.index, expected_boreholes_returned)
        assert_array_almost_equal(result["V"], expected_tops)

        result = borehole_data.get_layer_top("lith", ["Z", "K"])
        expected_boreholes_returned = ["A", "B", "C", "D", "E"]
        expected_sand_top = [1.5, np.nan, 2.9, 2.5, 0.0]
        expected_clay_top = [0.0, 0.0, 0.0, 0.0, np.nan]

        assert result.shape == (5, 2)
        assert_array_equal(result.index, expected_boreholes_returned)
        assert_array_almost_equal(result["Z"], expected_sand_top)
        assert_array_almost_equal(result["K"], expected_clay_top)

    @pytest.mark.unittest
    def test_slice_depth_interval(self, borehole_data):
        # Test slicing with respect to depth below the surface.
        upper, lower = 0.6, 2.4
        sliced = borehole_data.slice_depth_interval(upper, lower)
        assert isinstance(sliced, LayeredData)

        layers_per_borehole = sliced["nr"].value_counts()
        expected_layer_count = [3, 3, 3, 3, 2]

        assert len(sliced) == 14
        assert sliced["top"].min() == upper
        assert sliced["bottom"].max() == lower
        assert_array_equal(layers_per_borehole, expected_layer_count)

        # Test slicing without updating layer boundaries.
        sliced = borehole_data.slice_depth_interval(
            upper, lower, update_layer_boundaries=False
        )

        expected_tops_of_slice = [0.0, 0.6, 0.0, 0.5, 0.5]
        expected_bottoms_of_slice = [2.5, 2.5, 2.9, 2.5, 2.5]

        tops_of_slice = sliced.df.groupby("nr")["top"].min()
        bottoms_of_slice = sliced.df.groupby("nr")["bottom"].max()

        assert len(sliced) == 14
        assert_array_equal(tops_of_slice, expected_tops_of_slice)
        assert_array_equal(bottoms_of_slice, expected_bottoms_of_slice)

        # Test slicing with respect to a vertical reference plane.
        nap_upper, nap_lower = -2, -3
        sliced = borehole_data.slice_depth_interval(
            nap_upper, nap_lower, relative_to_vertical_reference=True
        )

        expected_tops_of_slice = [2.2, 2.3, 2.25, 2.1, 1.9]
        expected_bottoms_of_slice = [3.2, 3.3, 3.25, 3.0, 2.9]

        tops_of_slice = sliced.df.groupby("nr")["top"].min()
        bottoms_of_slice = sliced.df.groupby("nr")["bottom"].max()

        assert len(sliced) == 11
        assert_array_equal(tops_of_slice, expected_tops_of_slice)
        assert_array_equal(bottoms_of_slice, expected_bottoms_of_slice)

        # Test slices that return empty objects.
        empty_slice = borehole_data.slice_depth_interval(-2, -1)
        empty_slice_nap = borehole_data.slice_depth_interval(
            3, 2, relative_to_vertical_reference=True
        )

        assert len(empty_slice) == 0
        assert len(empty_slice_nap) == 0

        # Test slicing using only an upper boundary or lower boundary.
        upper = 4
        sliced = borehole_data.slice_depth_interval(upper)

        expected_boreholes = ["A", "C"]

        assert len(sliced) == 2
        assert_array_equal(sliced["nr"], expected_boreholes)

        nap_lower = -0.5
        sliced = borehole_data.slice_depth_interval(
            lower_boundary=nap_lower, relative_to_vertical_reference=True
        )

        bottoms_of_slice = sliced.df.groupby("nr")["bottom"].max()
        expected_bottoms_of_slice = [0.7, 0.8, 0.75, 0.6, 0.4]

        assert len(sliced) == 7
        assert_array_equal(bottoms_of_slice, expected_bottoms_of_slice)

    @pytest.mark.unittest
    def test_select_by_condition(self, borehole_data):
        selected = borehole_data.select_by_condition(borehole_data["lith"] == "V")
        assert isinstance(selected, LayeredData)

        expected_nrs = ["B", "D"]
        assert_array_equal(selected["nr"].unique(), expected_nrs)
        assert np.all(selected["lith"] == "V")
        assert len(selected) == 4

        selected = borehole_data.select_by_condition(
            borehole_data["lith"] == "V", invert=True
        )
        assert len(selected) == 21
        assert ~np.all(selected["lith"] == "V")

    @pytest.mark.unittest
    def test_to_pyvista_cylinders(self, borehole_data):
        # Test normal to multiblock.
        multiblock = borehole_data.to_pyvista_cylinders("lith")
        expected_bounds = (0.0, 5.0, 0.0, 6.0, -5.25, 0.3)
        assert isinstance(multiblock, MultiBlock)
        assert multiblock.n_blocks == 5
        assert multiblock.bounds == expected_bounds
        assert multiblock[0].n_arrays == 2
        assert multiblock[0].n_cells == 22
        assert multiblock[0].n_points == 160

        # Test with vertical exageration.
        multiblock = borehole_data.to_pyvista_cylinders("lith", vertical_factor=10)
        expected_bounds = (0.0, 5.0, 0.0, 6.0, -52.5, 3.0)
        assert multiblock.n_blocks == 5
        assert multiblock.bounds == expected_bounds

        # Test to multiblock with respect to depth below the surface.
        multiblock = borehole_data.to_pyvista_cylinders(
            "lith", relative_to_vertical_reference=False
        )
        expected_bounds = (0.0, 5.0, 0.0, 6.0, 0.0, 5.5)
        assert multiblock.n_blocks == 5
        assert multiblock.bounds == expected_bounds

        # Test with both options.
        multiblock = borehole_data.to_pyvista_cylinders(
            "lith", vertical_factor=10, relative_to_vertical_reference=False
        )
        expected_bounds = (0.0, 5.0, 0.0, 6.0, 0.0, 55.0)
        assert multiblock.n_blocks == 5
        assert multiblock.bounds == expected_bounds

    @pytest.mark.unittest
    def test_to_pyvista_grid(self, borehole_data):
        grid = borehole_data.to_pyvista_grid(["lith"], radius=0.1)
        assert isinstance(grid, UnstructuredGrid)
        assert grid.n_cells > 0
        assert "lith" in grid.array_names

    @pytest.mark.unittest
    def test_to_datafusiontools(self, borehole_data):
        # Test normal export.
        dft = borehole_data.to_datafusiontools("lith")

        expected_independent_value = [-0.2, -0.95, -1.8, -2.9, -3.75]
        expected_number_of_variables = 1

        assert len(dft) == 5
        assert np.all([isinstance(d, geodataclass.Data) for d in dft])
        assert np.all([len(d.variables) == expected_number_of_variables for d in dft])
        assert_array_almost_equal(
            dft[0].independent_variable.value, expected_independent_value
        )

        # Test with label encoding.
        dft = borehole_data.to_datafusiontools("lith", encode=True)
        expected_number_of_variables = 3
        assert np.all([isinstance(d, geodataclass.Data) for d in dft])
        assert np.all([len(d.variables) == expected_number_of_variables for d in dft])

        # Test without updating layer depths to NAP
        dft = borehole_data.to_datafusiontools(
            "lith", relative_to_vertical_reference=False
        )
        expected_independent_value = [0.4, 1.15, 2.0, 3.1, 3.95]

        assert_array_almost_equal(
            dft[0].independent_variable.value, expected_independent_value
        )

    @pytest.mark.unittest
    def test_to_datafusiontools_with_file(self, borehole_data):
        outfile = Path("dft.pkl")
        borehole_data.to_datafusiontools("lith", outfile)
        assert outfile.is_file()
        outfile.unlink()

    @pytest.mark.unittest
    def test_to_qgis3d(self, borehole_data):
        outfile = Path("temp.gpkg")
        borehole_data.to_qgis3d(outfile, crs=28992)
        assert outfile.is_file()
        outfile.unlink()

    @pytest.mark.unittest
    def test_to_kingdom(self, borehole_data):
        outfile = Path("temp_kingdom.csv")
        tdfile = Path(outfile.parent, f"{outfile.stem}_TDCHART{outfile.suffix}")
        borehole_data.to_kingdom(outfile)
        assert outfile.is_file()
        assert tdfile.is_file()
        out_layerdata = pd.read_csv(outfile)
        out_tddata = pd.read_csv(tdfile)
        assert_array_almost_equal(
            out_layerdata["Start depth"][:6],
            np.array([0.0, 0.8, 1.5, 2.5, 3.7, 0.0]),
        )
        assert_array_almost_equal(
            out_layerdata["End depth"][:6],
            np.array([0.8, 1.5, 2.5, 3.7, 4.2, 0.6]),
        )
        assert_array_almost_equal(
            out_layerdata["Total depth"][:6],
            np.array([4.2, 4.2, 4.2, 4.2, 4.2, 3.9]),
        )
        assert_array_almost_equal(
            out_tddata["MD"],
            np.array([0, 1, 0, 1, 0, 1, 0, 1, 0, 1]),
        )
        assert_array_almost_equal(
            out_tddata["TWT"],
            np.array(
                [
                    -0.26666667,
                    0.98333333,
                    -0.4,
                    0.85,
                    -0.33333333,
                    0.91666667,
                    -0.13333333,
                    1.11666667,
                    0.13333333,
                    1.38333333,
                ]
            ),
        )
        outfile.unlink()
        tdfile.unlink()

    @pytest.mark.unittest
    def test_create_geodataframe_3d(self, borehole_data):
        relative_to_vertical_reference = True
        gdf = borehole_data._create_geodataframe_3d(
            relative_to_vertical_reference, crs=28992
        )

        first_line_coords = get_coordinates(gdf["geometry"].iloc[0], include_z=True)
        expected_coords = [[2.0, 3.0, 0.21], [2.0, 3.0, -0.59]]

        assert all(gdf.geom_type == "LineString")
        assert_array_almost_equal(first_line_coords, expected_coords)

        relative_to_vertical_reference = False
        gdf = borehole_data._create_geodataframe_3d(
            relative_to_vertical_reference, crs=28992
        )

        first_line_coords = get_coordinates(gdf["geometry"].iloc[0], include_z=True)
        expected_coords = [[2.0, 3.0, 0.01], [2.0, 3.0, 0.81]]

        assert all(gdf.geom_type == "LineString")
        assert_array_almost_equal(first_line_coords, expected_coords)

    @pytest.mark.unittest
    def test_change_depth_values(self, borehole_data):
        borehole = borehole_data.select_by_values("nr", "A").df
        borehole = borehole_data._change_depth_values(borehole)

        expected_top = [0.2, -0.6, -1.3, -2.3, -3.5]
        expected_bottom = [-0.6, -1.3, -2.3, -3.5, -4.0]

        assert_array_almost_equal(borehole["top"], expected_top)
        assert_array_almost_equal(borehole["bottom"], expected_bottom)

    @pytest.mark.unittest
    def test_check_correct_instance(self, borehole_data):
        inst = "string"
        inst = borehole_data._check_correct_instance(inst)
        assert isinstance(inst, list)

        inst = ["list of strings"]
        inst = borehole_data._check_correct_instance(inst)
        assert isinstance(inst, list)

    @pytest.mark.unittest
    def test_to_csv_mixin(self, borehole_data):
        outfile = Path("temp.csv")
        borehole_data.to_csv(outfile)
        assert outfile.is_file()
        outfile.unlink()

    @pytest.mark.unittest
    def test_to_parquet_mixin(self, borehole_data):
        outfile = Path("temp.parquet")
        borehole_data.to_parquet(outfile)
        assert outfile.is_file()
        outfile.unlink()


class TestDiscreteData:
    @pytest.mark.unittest
    def test_datatype(self, cpt_data):
        assert cpt_data.datatype == "discrete"

    @pytest.mark.unittest
    def test_to_header(self, cpt_data):
        header = cpt_data.to_header()
        expected_columns = ["nr", "x", "y", "surface", "end", "geometry"]

        header = cpt_data.to_header()

        assert isinstance(header, PointHeader)
        assert_array_equal(header.gdf.columns, expected_columns)
        assert len(header.gdf) == 2
        assert header["nr"].nunique() == 2
        assert header.horizontal_reference == 28992
        assert header.vertical_reference == 5709

    @pytest.mark.unittest
    def test_to_collection(self, cpt_data):
        collection = cpt_data.to_collection()
        assert isinstance(collection, CptCollection)
        assert isinstance(collection.header, PointHeader)
        assert len(collection.header) == 2

    @pytest.mark.unittest
    def test_select_by_values(self, cpt_data):
        selected = cpt_data.select_by_values("nr", "a")
        assert isinstance(selected, DiscreteData)
        assert len(selected) == 10
        assert_array_equal(selected["nr"].unique(), "a")

    @pytest.mark.unittest
    def test_select_by_condition(self, cpt_data):
        selected = cpt_data.select_by_condition(
            (cpt_data["qc"] > 0.3) & (cpt_data["qc"] < 0.4)
        )
        assert isinstance(selected, DiscreteData)
        assert len(selected) == 5
        assert np.all((selected["qc"] > 0.3) & (selected["qc"] < 0.4))
        assert_array_equal(selected.df.index, [2, 3, 4, 5, 6])

        selected = cpt_data.select_by_condition(
            (cpt_data["qc"] < 0.35) & (cpt_data["fs"] < 0.2)
        )
        assert len(selected) == 3
        assert np.all(selected["qc"] < 0.35)
        assert np.all(selected["fs"] < 0.2)
        assert_array_equal(selected.df.index, [0, 1, 2])

    @pytest.mark.unittest
    def test_slice_depth_interval(self, cpt_data):
        # Selection with respect to surface level
        selected = cpt_data.slice_depth_interval(2, 3)
        assert isinstance(selected, DiscreteData)
        assert len(selected) == 4
        assert_array_equal(selected["depth"], [2, 3, 2, 3])

        # Selection with respect to vertical reference plane
        selected = cpt_data.slice_depth_interval(
            1.9, 0.9, relative_to_vertical_reference=True
        )
        assert len(selected) == 1
        assert_array_equal(selected["depth"], [1])
        assert_array_equal(selected["nr"], ["a"])

        # Selection with respect to surface level using one limit
        selected = cpt_data.slice_depth_interval(lower_boundary=0.9)
        assert len(selected) == 2
        assert_array_equal(selected["depth"], [0, 0])
        assert_array_equal(selected["nr"], ["a", "b"])

        # Selection with respect to vertical reference plane using one limit
        selected = cpt_data.slice_depth_interval(
            lower_boundary=0.9, relative_to_vertical_reference=True
        )
        assert len(selected) == 2
        assert_array_equal(selected["depth"], [0, 1])
        assert np.all(selected["nr"] == "a")


class TestDataclasses:
    @pytest.mark.unittest
    def test_cpt_dataclass(self):
        cpt = Cpt()
        assert isinstance(cpt, Cpt)
        assert cpt.nr is None
        assert cpt.x is None
        assert cpt.y is None
        assert cpt.z is None
        assert cpt.enddepth is None
        assert cpt.gefid is None
        assert cpt.ncolumns is None
        assert cpt.columninfo is None
        assert cpt.companyid is None
        assert cpt.filedate is None
        assert cpt.fileowner is None
        assert cpt.lastscan is None
        assert cpt.procedurecode is None
        assert cpt.reportcode is None
        assert cpt.projectid is None



import sqlite3

import geopandas as gpd
import pandas as pd
import pytest
from pyogrio.errors import FieldError
from shapely import geometry as gmt

from geost import utils


@pytest.mark.unittest
def test_save_pickle(borehole_collection, tmp_path):
    outfile = tmp_path / r"collection.pkl"
    utils.save_pickle(borehole_collection, outfile)
    assert outfile.is_file()


@pytest.mark.unittest
def test_to_geopackage(borehole_collection, tmp_path):
    outfile = tmp_path / "gpkg_with_error.gpkg"
    invalid_column = "GEOM"
    borehole_collection.header[invalid_column] = "test"
    with pytest.raises(FieldError):
        borehole_collection.to_geopackage(outfile)


@pytest.mark.unittest
def test_create_connection(simple_soilmap_gpkg):
    conn = utils.create_connection(simple_soilmap_gpkg)
    assert isinstance(conn, sqlite3.Connection)


@pytest.mark.unittest
@pytest.mark.parametrize(
    "input_value,expected",
    [
        ("3.14", 3.14),
        ("0", 0.0),
        ("-2.5", -2.5),
        ("1e3", 1000.0),
        ("not_a_number", None),
        ("", None),
        (5, 5.0),
        (3.0, 3.0),
    ],
)
def test_safe_float(input_value, expected):
    result = utils.safe_float(input_value)
    assert (result == expected) or (result is None and expected is None)


@pytest.mark.unittest
def test_safe_float_invalid_type():
    """Test safe_float with an invalid type."""
    with pytest.raises(TypeError):
        utils.safe_float([])


@pytest.mark.unittest
def test_get_path_iterable_file(tmp_path):
    file = tmp_path / "test.csv"
    file.write_text("a,b\n1,2")
    result = utils.get_path_iterable(file)
    assert isinstance(result, list)
    assert result == [file]


@pytest.mark.unittest
def test_get_path_iterable_dir(tmp_path):
    (tmp_path / "a.txt").write_text("foo")
    (tmp_path / "b.txt").write_text("bar")
    result = list(utils.get_path_iterable(tmp_path, "*.txt"))
    assert len(result) == 2
    assert all(f.suffix == ".txt" for f in result)


@pytest.mark.unittest
def test_get_path_iterable_invalid(tmp_path):
    invalid_path = tmp_path / "does_not_exist"
    with pytest.raises(TypeError):
        utils.get_path_iterable(invalid_path)


@pytest.mark.parametrize(
    "geometry, expected_length",
    [
        (gpd.GeoDataFrame(geometry=[gmt.Point(1, 2)]), 1),
        (gmt.Point(1, 2), 1),
        (gmt.LineString([(0, 0), (1, 1)]), 1),
        (gmt.Polygon([(0, 0), (1, 1), (1, 0), (0, 0)]), 1),
        (gmt.MultiPoint([gmt.Point(1, 2), gmt.Point(3, 4)]), 1),
        ([gmt.Point(1, 2), gmt.Point(3, 4)], 2),
    ],
)
def test_check_geometry_instance(geometry, expected_length):
    geom = utils.check_geometry_instance(geometry)
    assert isinstance(geom, gpd.GeoDataFrame)
    assert len(geom) == expected_length


@pytest.mark.unittest
def test_check_geometry_instance_from_file(tmp_path, point_header_gdf):
    outfile = tmp_path / "test.geoparquet"
    point_header_gdf.to_parquet(outfile)
    gdf = utils.check_geometry_instance(outfile)
    assert isinstance(gdf, gpd.GeoDataFrame)


@pytest.mark.parametrize("extension", [".geoparquet", ".parquet", ".gpkg", ".shp"])
def test_geopandas_read(tmp_path, point_header_gdf, extension):
    outfile = tmp_path / f"test{extension}"

    if extension in {".geoparquet", ".parquet"}:
        point_header_gdf.to_parquet(outfile)
    elif extension in {".gpkg", ".shp"}:
        point_header_gdf.to_file(outfile)

    gdf = utils._geopandas_read(outfile)
    assert isinstance(gdf, gpd.GeoDataFrame)


@pytest.mark.unittest
def test_geopandas_read_invalid_file(tmp_path):
    invalid_file = tmp_path / "invalid.txt"
    invalid_file.write_text("This is not a valid geopandas file.")

    with pytest.raises(
        ValueError, match="File type .txt is not supported by geopandas."
    ):
        utils._geopandas_read(invalid_file)


@pytest.mark.parametrize(
    "extension", [".pq", ".parquet", ".csv", ".txt", ".tsv", ".xls", ".xlsx"]
)
def test_pandas_read(tmp_path, point_header_gdf, extension):
    outfile = tmp_path / f"test{extension}"

    if extension in {".pq", ".parquet"}:
        point_header_gdf.to_parquet(outfile)
    elif extension in {".csv", ".txt", ".tsv"}:
        point_header_gdf.to_csv(outfile)
    elif extension in {".xls", ".xlsx"}:
        point_header_gdf.to_excel(outfile)

    df = utils._pandas_read(outfile)
    assert isinstance(df, pd.DataFrame)



from unittest.mock import Mock

import geopandas as gpd
import numpy as np
import pytest
from shapely.geometry import LineString, Point

from geost.base import LineHeader, PointHeader
from geost.header_factory import HeaderFactory, header_factory


class TestHeaderFactory:
    """Test suite for HeaderFactory class."""

    @pytest.fixture
    def factory(self):
        """Create a fresh HeaderFactory instance for each test."""
        return HeaderFactory()

    @pytest.fixture
    def point_gdf(self):
        """Create a simple point GeoDataFrame for testing."""
        data = {
            "nr": ["test1", "test2"],
            "x": [1.0, 2.0],
            "y": [1.0, 2.0],
            "surface": [0.0, 0.0],
            "end": [-10.0, -10.0],
        }
        geometry = [Point(1.0, 1.0), Point(2.0, 2.0)]
        gdf = gpd.GeoDataFrame(data, geometry=geometry, crs="EPSG:28992")
        return gdf

    @pytest.fixture
    def line_gdf(self):
        """Create a simple line GeoDataFrame for testing."""
        data = {
            "nr": ["line1", "line2"],
            "x": [1.0, 2.0],
            "y": [1.0, 2.0],
        }
        geometry = [LineString([(0, 0), (1, 1)]), LineString([(1, 1), (2, 2)])]
        gdf = gpd.GeoDataFrame(data, geometry=geometry, crs="EPSG:28992")
        return gdf

    @pytest.fixture
    def mixed_geometry_gdf(self):
        """Create a GeoDataFrame with mixed geometry types."""
        data = {
            "nr": ["test1", "test2"],
            "x": [1.0, 2.0],
            "y": [1.0, 2.0],
        }
        geometry = [Point(1.0, 1.0), LineString([(0, 0), (1, 1)])]
        gdf = gpd.GeoDataFrame(data, geometry=geometry, crs="EPSG:28992")
        return gdf

    @pytest.mark.unittest
    def test_init(self, factory):
        """Test HeaderFactory initialization."""
        assert isinstance(factory._types, dict)
        assert len(factory._types) == 0

    @pytest.mark.unittest
    def test_register_header(self, factory):
        """Test registering header types."""
        # Test registering Point header
        factory.register_header("Point", PointHeader)
        assert "Point" in factory._types
        assert factory._types["Point"] == PointHeader

        # Test registering Line header
        factory.register_header("Line", LineHeader)
        assert "Line" in factory._types
        assert factory._types["Line"] == LineHeader

        # Test overwriting existing registration
        mock_header = Mock()
        factory.register_header("Point", mock_header)
        assert factory._types["Point"] == mock_header

    @pytest.mark.unittest
    def test_get_geometry_type_single_point(self, point_gdf):
        """Test getting geometry type for Point geometries."""
        geometry_type = HeaderFactory.get_geometry_type(point_gdf)
        assert geometry_type == "Point"

    @pytest.mark.unittest
    def test_get_geometry_type_single_line(self, line_gdf):
        """Test getting geometry type for LineString geometries."""
        geometry_type = HeaderFactory.get_geometry_type(line_gdf)
        assert geometry_type == "LineString"

    @pytest.mark.unittest
    def test_get_geometry_type_mixed(self, mixed_geometry_gdf):
        """Test getting geometry type for mixed geometries."""
        geometry_type = HeaderFactory.get_geometry_type(mixed_geometry_gdf)
        assert geometry_type == "Multiple geometries"

    @pytest.mark.unittest
    def test_get_geometry_type_empty(self):
        """Test getting geometry type for empty GeoDataFrame."""
        empty_gdf = gpd.GeoDataFrame(
            {"nr": [], "x": [], "y": []}, geometry=[], crs="EPSG:28992"
        )
        geometry_type = HeaderFactory.get_geometry_type(empty_gdf)
        assert geometry_type is None

    @pytest.mark.unittest
    def test_create_from_point_success(self, factory, point_gdf):
        """Test successful creation of PointHeader."""
        factory.register_header("Point", PointHeader)

        header = factory.create_from(point_gdf, vertical_reference="NAP")

        assert isinstance(header, PointHeader)
        assert len(header) == 2

    @pytest.mark.unittest
    def test_create_from_line_success(self, factory, line_gdf):
        """Test successful creation of LineHeader."""
        factory.register_header("LineString", LineHeader)

        header = factory.create_from(line_gdf, vertical_reference="NAP")

        assert isinstance(header, LineHeader)
        assert len(header) == 2

    @pytest.mark.unittest
    def test_create_from_unregistered_type(self, factory, point_gdf):
        """Test creation with unregistered geometry type raises ValueError."""
        # Don't register Point header
        with pytest.raises(ValueError, match="No Header type available for Point"):
            factory.create_from(point_gdf, vertical_reference="NAP")

    @pytest.mark.unittest
    def test_create_from_mixed_geometries(self, factory, mixed_geometry_gdf):
        """Test creation with mixed geometries raises ValueError."""
        factory.register_header("Point", PointHeader)
        factory.register_header("LineString", LineHeader)

        with pytest.raises(
            ValueError, match="No Header type available for Multiple geometries"
        ):
            factory.create_from(mixed_geometry_gdf, vertical_reference="NAP")

    @pytest.mark.unittest
    def test_create_from_with_kwargs(self, factory, point_gdf):
        """Test that kwargs are passed through to header constructor."""
        factory.register_header("Point", PointHeader)

        # Test with vertical_reference kwarg
        header = factory.create_from(point_gdf, vertical_reference="EPSG:5709")
        assert header.vertical_reference.to_epsg() == 5709

    @pytest.mark.unittest
    def test_create_from_empty_gdf(self, factory):
        """Test creation with empty GeoDataFrame."""
        factory.register_header("Point", PointHeader)

        empty_gdf = gpd.GeoDataFrame(
            {"nr": [], "x": [], "y": [], "surface": [], "end": []},
            geometry=[],
            crs="EPSG:28992",
        )

        # Empty GeoDataFrame should raise ValueError due to None geometry type
        with pytest.raises(ValueError, match="No Header type available for None"):
            factory.create_from(empty_gdf, vertical_reference="NAP")



import sqlite3

import pandas as pd
import pytest
from numpy.testing import assert_array_equal

from geost.io import Geopackage


class TestGeopackage:
    @pytest.mark.unittest
    def test_get_connection(self, simple_soilmap_gpkg):
        gp = Geopackage(simple_soilmap_gpkg)
        gp.get_connection()
        assert isinstance(gp.connection, sqlite3.Connection)

    @pytest.mark.unittest
    def test_layers(self, simple_soilmap_gpkg):
        gp = Geopackage(simple_soilmap_gpkg)
        layers = gp.layers()
        assert isinstance(layers, pd.DataFrame)
        assert_array_equal(layers["name"], ["soilarea", "soilarea_soilunit"])
        assert_array_equal(layers["geometry_type"], ["Polygon", None])

    @pytest.mark.unittest
    def test_context_manager(self, simple_soilmap_gpkg):
        gp = Geopackage(simple_soilmap_gpkg)
        assert gp.connection is None
        with gp:
            assert isinstance(gp.connection, sqlite3.Connection)
        assert gp.connection is None

    @pytest.mark.unittest
    def test_get_cursor(self, simple_soilmap_gpkg):
        with Geopackage(simple_soilmap_gpkg) as gp:
            cursor = gp._get_cursor()
            assert isinstance(cursor, sqlite3.Cursor)

    @pytest.mark.unittest
    def test_get_column_names(self, simple_soilmap_gpkg):
        with Geopackage(simple_soilmap_gpkg) as gp:
            columns = gp.get_column_names("soilarea_soilunit")
            assert_array_equal(columns, ["fid", "maparea_id", "soilunit_code"])

    @pytest.mark.unittest
    def test_read_table(self, simple_soilmap_gpkg):
        test_table = "soilarea_soilunit"
        with Geopackage(simple_soilmap_gpkg) as gp:
            table = gp.read_table(test_table)
            assert isinstance(table, pd.DataFrame)
            assert_array_equal(table.columns, ["fid", "maparea_id", "soilunit_code"])

            table = gp.table_head(test_table)
            assert len(table) == 5

    @pytest.mark.unittest
    def test_query(self, simple_soilmap_gpkg):
        with Geopackage(simple_soilmap_gpkg) as gp:
            query = "SELECT * FROM soilarea_soilunit"
            table = gp.query(query)
        assert isinstance(table, pd.DataFrame)
        assert_array_equal(table.columns, ["fid", "maparea_id", "soilunit_code"])

        with Geopackage(simple_soilmap_gpkg) as gp:
            query = "SELECT * FROM soilarea_soilunit WHERE fid = 1"
            table = gp.query(query, outcolumns=["A", "B", "C"])
        assert_array_equal(table.columns, ["A", "B", "C"])



import numpy as np

# NOTE: temporary functions ripped from old code. To be made compatible with Geolib
# functions


def calc_ic(qc, rf) -> np.ndarray:
    """
    Calculate non-normalized IC values (I_SBT in Robertson 2010). The non-normalized
    variant does not require calculations of stresses to normalize with and hence no PWP
    data is required.

    Please note the following when using non-normalized IC values:

    "The non-normalized SBT index (ISBT) is essentially the same as the normalized SBTn
    index (Ic) but only uses the basic CPT measurements. In general, the normalized Ic
    provides more reliable identification of SBT than the non-normalized ISBT, but when
    the insitu vertical effective stress is between 50 kPa to 150 kPa there is often
    little difference between normalized and non-normalized SBT."

    Parameters
    ----------
    qc : np.ndarray
        Cone resistance values
    rf : np.ndarray
        Friction number

    Returns
    -------
    np.ndarray
        Non-normalized IC
    """
    ic = np.sqrt((3.47 - np.log10(qc / 0.1)) ** 2 + (np.log10(rf) + 1.22) ** 2)
    return ic


# TODO numpy searchsort
def calc_lithology(ic, qc, rf) -> np.ndarray:
    boundaries = [1.6, 2.0, 2.2, 2.6, 2.95, 3.6]
    lith = np.full_like(ic, "NBE", dtype="<U3")
    lith[ic < boundaries[0]] = "Z"
    lith[(ic >= boundaries[0]) & (ic < boundaries[1])] = "Z"
    lith[(ic >= boundaries[1]) & (ic < boundaries[2])] = "Z"
    lith[(ic >= boundaries[2]) & (ic < boundaries[3])] = "Z"
    lith[(ic >= boundaries[3]) & (ic < boundaries[4])] = "Kz"
    lith[(ic >= boundaries[4]) & (ic < boundaries[5])] = "K"
    lith[(ic >= boundaries[5]) & (rf > 8)] = "V"
    lith[((rf > 5) & (qc < 1.5)) | (rf > 6)] = "V"
    lith[(ic >= boundaries[5]) & (rf <= 8)] = "Kh"
    return lith



import warnings
from pathlib import Path
from typing import Iterable

import geopandas as gpd
import numpy as np
import pandas as pd
import rioxarray
import xarray as xr

from geost import utils


def check_and_coerce_crs(gdf: gpd.GeoDataFrame, to_crs: int):
    """
    Check the CRS of a geodataframe against given crs.

    If the geodataframe has no CRS, makes the geodataframe assume the given CRS. The
    user is warned when this occurs.

    If the geodataframe has a different known CRS, inform user and coerce the CRS
    to the given CRS (to_crs argument).

    Parameters
    ----------
    gdf : GeoDataFrame
        Geodataframe to be converted to the given CRS.
    to_crs : int
        EPSG number of the CRS to check and coerce to.

    Returns
    -------
    GeoDataFrame
        Geodataframe coerced to have the desired CRS
    """
    if gdf.crs is None:
        gdf.crs = to_crs
        warnings.warn(
            "The selection geometry has no crs! Assuming it is the same as the "
            + f"horizontal_reference (epsg:{to_crs}) of this "
            + "collection. PLEASE CHECK WHETHER THIS IS CORRECT!",
            UserWarning,
        )
    elif gdf.crs != to_crs:
        gdf = gdf.to_crs(to_crs)
        warnings.warn(
            "The crs of the selection geometry does not match the horizontal "
            + "reference of the collection. The selection geometry was coerced "
            + f"to epsg:{to_crs} automatically",
            UserWarning,
        )
    return gdf


def select_points_within_bbox(
    gdf: str | Path | gpd.GeoDataFrame,
    xmin: float | int,
    ymin: float | int,
    xmax: float | int,
    ymax: float | int,
    invert: bool = False,
) -> gpd.GeoDataFrame:
    """
    Make a selection of point geometries based on a user-given bounding box.

    Parameters
    ----------
    gdf : str | Path | gpd.GeoDataFrame
        Geodataframe (or file that can be parsed to a geodataframe) to select from.
    xmin : float | int
        Minimum x-coordinate of the bounding box.
    ymin : float | int
        Minimum y-coordinate of the bounding box.
    xmax : float | int
        Maximum x-coordinate of the bounding box.
    ymax : float | int
        Maximum y-coordinate of the bounding box.
    invert : bool, optional
        Invert the selection, so select all objects outside of the
        bounding box in this case, by default False.

    Returns
    -------
    gpd.GeoDataFrame
        Geodataframe containing only selected geometries.
    """
    # Instance checks and coerce to geodataframe if required
    gdf = utils.check_geometry_instance(gdf)

    # Selection logic
    x = gdf["geometry"].x
    y = gdf["geometry"].y
    if invert:
        gdf_selected = gdf[(x < xmin) | (x > xmax) | (y < ymin) | (y > ymax)]
    else:
        gdf_selected = gdf[(x >= xmin) & (x <= xmax) & (y >= ymin) & (y <= ymax)]
    return gdf_selected


def select_points_near_points(
    gdf: str | Path | gpd.GeoDataFrame,
    point_gdf: str | Path | gpd.GeoDataFrame,
    buffer: float | int,
    invert: bool = False,
) -> gpd.GeoDataFrame:
    """
    Make a selection of point geometries based on point geometries and a buffer.

    Parameters
    ----------
    gdf : str | Path | gpd.GeoDataFrame
        Geodataframe (or file that can be parsed to a geodataframe) to select from.
    point_gdf : str | Path | gpd.GeoDataFrame
        Geodataframe (or file that can be parsed to a geodataframe) to select with.
    buffer : float | int
        Buffer distance for selection geometries.
    invert : bool, optional
        Invert the selection, by default False.

    Returns
    -------
    gpd.GeoDataFrame
        Geodataframe containing only selected geometries.
    """
    # Instance checks and coerce to geodataframe if required
    gdf = utils.check_geometry_instance(gdf)
    point_gdf = utils.check_geometry_instance(point_gdf)

    # Selection logic
    data_points = np.array([gdf["geometry"].x, gdf["geometry"].y]).transpose()
    query_points = np.array(
        [point_gdf["geometry"].x, point_gdf["geometry"].y]
    ).transpose()

    bool_array = np.full(len(data_points), False)
    for query_point in query_points:
        distance = np.sqrt(
            (query_point[0] - data_points[:, 0]) ** 2
            + (query_point[1] - data_points[:, 1]) ** 2
        )
        bool_array += distance < buffer
    if invert:
        bool_array = np.invert(bool_array)
    gdf_selected = gdf[bool_array]
    return gdf_selected


def select_points_near_lines(
    gdf: str | Path | gpd.GeoDataFrame,
    line_gdf: str | Path | gpd.GeoDataFrame,
    buffer: float | int,
    invert: bool = False,
) -> gpd.GeoDataFrame:
    """
    Make a selection of point geometries based on line geometries and a buffer.

    Parameters
    ----------
    gdf : str | Path | gpd.GeoDataFrame
        Geodataframe (or file that can be parsed to a geodataframe) to select from.
    line_gdf : str | Path | gpd.GeoDataFrame
        Geodataframe (or file that can be parsed to a geodataframe) to select with.
    buffer : float | int
        Buffer distance for selection geometries.
    invert : bool, optional
        Invert the selection, by default False.

    Returns
    -------
    gpd.GeoDataFrame
        Geodataframe containing only selected geometries.
    """
    # Instance checks and coerce to geodataframe if required
    gdf = utils.check_geometry_instance(gdf)
    line_gdf = utils.check_geometry_instance(line_gdf)

    # Selection logic
    line_gdf["geometry"] = line_gdf.buffer(distance=buffer)
    if invert:
        gdf_selected = gdf[~gdf.geometry.within(line_gdf.union_all())]
    else:
        gdf_selected = gdf[gdf.geometry.within(line_gdf.union_all())]
    return gdf_selected


def select_points_within_polygons(
    gdf: str | Path | gpd.GeoDataFrame,
    polygon_gdf: str | Path | gpd.GeoDataFrame,
    buffer: float | int = 0,
    invert: bool = False,
) -> gpd.GeoDataFrame:
    """
    Make a selection of point geometries based on polygon geometries and an optional
    buffer.

    Parameters
    ----------
    gdf : str | Path | gpd.GeoDataFrame
        Geodataframe (or file that can be parsed to a geodataframe) to select from.
    polygon_gdf : str | Path | gpd.GeoDataFrame
        Geodataframe (or file that can be parsed to a geodataframe) to select with.
    buffer : float | int, optional
        Optional buffer distance around the polygon selection geometries, by default 0.
    invert : bool, optional
        Invert the selection, by default False.

    Returns
    -------
    gpd.GeoDataFrame
        Geodataframe containing only selected geometries.
    """
    # Instance checks and coerce to geodataframe if required
    gdf = utils.check_geometry_instance(gdf)
    polygon_gdf = utils.check_geometry_instance(polygon_gdf)

    # Selection logic
    if buffer > 0:
        polygon_select = polygon_gdf.copy()
        polygon_select["geometry"] = polygon_gdf.geometry.buffer(buffer)
    else:
        polygon_select = polygon_gdf

    if invert:
        gdf_selected = gdf[~gdf.geometry.within(polygon_select.union_all())]
    else:
        gdf_selected = gdf[gdf.geometry.within(polygon_select.union_all())]

    return gdf_selected


def find_area_labels(
    point_geodataframe: gpd.GeoDataFrame,
    polygon_geodataframe: gpd.GeoDataFrame,
    column_name: str | Iterable,
) -> pd.Series:
    """
    Function to find labels associated with polygon geometries for a series of queried
    point geometries. Basically a spatial join between the point and polygon dataframe.

    Parameters
    ----------
    point_geodataframe : gpd.GeoDataFrame
        Geodataframe with point geometries for which you want to find in which polygon
        geometries they are located.
    polygon_geodataframe : gpd.GeoDataFrame
        Geodataframe with polygon geometries
    column_name : str | Iterable
        Label of the polygon geometries to use for assigning to the queried points.
        Given as a string or iterable of strings in case you'd like to find multiple
        labels.

    Returns
    -------
    pandas.Series
        Series with labels from the polygon geometries for each point.
    """
    if not isinstance(column_name, str):
        column_name = list(column_name)
    joined = gpd.sjoin(point_geodataframe, polygon_geodataframe)[column_name]
    # Remove any duplicated indices, which may sometimes happen
    area_labels = joined[~joined.index.duplicated()]
    return area_labels


def get_raster_values(
    x: np.ndarray, y: np.ndarray, raster_to_read: str | Path | xr.DataArray
) -> np.ndarray:
    """
    Return sampled values from a raster at the given (x, y) locations.

    Parameters
    ----------
    x : np.ndarray
        1D array of x-coordinates, same length as 'y'.
    y : np.ndarray
        1D array of y-coordinates, same length as 'x'.
    raster_to_read : str | Path | xr.DataArray
        Location of a raster file or an xr.DataArray with dimensions 'x' and 'y'. This
        raster is used to sample values from at all (x, y) locations.

    Returns
    -------
    np.ndarray
        1D array of sampled values
    """
    if isinstance(raster_to_read, (str, Path)):
        raster_to_read = rioxarray.open_rasterio(raster_to_read).squeeze()

    if set(raster_to_read.dims) != set(("y", "x")):
        raise TypeError(
            "The xr.DataArray to sample from does not have the "
            + "required 'x' and 'y' dimensions"
        )

    xmin, ymin, xmax, ymax = raster_to_read.rio.bounds()
    outside_x = (x < xmin) | (x > xmax)
    outside_y = (y < ymin) | (y > ymax)

    surface_levels = raster_to_read.sel(
        x=xr.DataArray(x, dims=("loc")),
        y=xr.DataArray(y, dims=("loc")),
        method="nearest",
    ).values

    surface_levels[outside_x | outside_y] = np.nan

    return surface_levels



from pathlib import Path

import geopandas as gpd
import pandas as pd


class PandasExportMixin:
    def to_csv(self, file: str | Path, **kwargs):
        """
        Write data to csv file.

        Parameters
        ----------
        file : str | Path
            Path to csv file to be written.
        **kwargs
            pd.DataFrame.to_csv kwargs. See relevant Pandas documentation.
        """
        self.df.to_csv(file, **kwargs)

    def to_parquet(self, file: str | Path, **kwargs):
        """
        Write data to parquet file.

        Parameters
        ----------
        file : str | Path
            Path to parquet file to be written.
        **kwargs
            pd.DataFrame.to_parquet kwargs. See relevant Pandas documentation.
        """
        self.df.to_parquet(file, **kwargs)


class GeopandasExportMixin:
    def to_csv(self, file: str | Path, **kwargs):
        """
        Write header to csv file as plain table (no geometries).

        Parameters
        ----------
        file : str | Path
            Path to csv file to be written.
        **kwargs
            pd.DataFrame.to_csv kwargs. See relevant Pandas documentation.
        """
        df = pd.DataFrame(self.gdf.drop(columns="geometry"))
        df.to_csv(file, **kwargs)

    def to_parquet(self, file: str | Path, **kwargs):
        """
        Write header to parquet file as plain table (no geometries).

        Parameters
        ----------
        file : str | Path
            Path to parquet file to be written.
        **kwargs
            pd.DataFrame.to_parquet kwargs. See relevant Pandas documentation.
        """
        df = pd.DataFrame(self.gdf.drop(columns="geometry"))
        df.to_parquet(file, **kwargs)

    def to_shape(self, file: str | Path, **kwargs):
        """
        Write header data to shapefile. You can use the resulting file to display
        borehole locations in GIS for instance.

        Parameters
        ----------
        file : str | Path
            Path to shapefile to be written.
        **kwargs
            gpd.GeoDataFrame.to_file kwargs. See relevant GeoPandas documentation.
        """
        self.gdf.to_file(file, **kwargs)

    def to_geopackage(self, file: str | Path, **kwargs):
        """
        Write header data to geopackage. You can use the resulting file to display
        borehole locations in GIS for instance.

        Parameters
        ----------
        file : str | Path
            Path to geopackage to be written.
        **kwargs
            gpd.GeoDataFrame.to_file kwargs. See relevant GeoPandas documentation.
        """
        self.gdf.to_file(file, **kwargs)

    def to_geoparquet(self, file: str | Path, **kwargs):
        """
        Write header data to geoparquet. You can use the resulting file to display
        borehole locations in GIS for instance. Please note that Geoparquet is supported
        by GDAL >= 3.5. For Qgis this means QGis >= 3.26

        Parameters
        ----------
        file : str | Path
            Path to shapefile to be written.
        **kwargs
            gpd.GeoDataFrame.to_parquet kwargs. See relevant Pandas documentation.
        """
        self.gdf.to_parquet(file, **kwargs)



import pytest
from numpy.testing import assert_array_equal

from geost.bro import Lithology


class TestUnitEnum:
    @pytest.mark.unittest
    def test_select_units(self):
        sel = Lithology.select_units(["clay", "fine_sand"])
        assert_array_equal(sel, [Lithology.clay, Lithology.fine_sand])
        assert_array_equal(sel, [2, 5])

        sel = Lithology.select_units("organic")
        assert_array_equal(sel, [Lithology.organic])
        assert_array_equal(sel, [1])

    @pytest.mark.unittest
    def test_select_values(self):
        sel = Lithology.select_values([2, 5])
        assert_array_equal(sel, [Lithology.clay, Lithology.fine_sand])
        assert_array_equal(sel, [2, 5])

        sel = Lithology.select_values(1)
        assert_array_equal(sel, [Lithology.organic])
        assert_array_equal(sel, [1])

    @pytest.mark.unittest
    def test_select_units_empty(self):
        sel = Lithology.select_units(["foo"])
        assert not sel

    @pytest.mark.unittest
    def test_select_values_empty(self):
        sel = Lithology.select_values(9999)
        assert not sel

    @pytest.mark.unittest
    def test_to_dict(self):
        d = Lithology.to_dict(key="unit")
        assert d == {
            "anthropogenic": 0,
            "organic": 1,
            "clay": 2,
            "loam": 3,
            "fine_sand": 5,
            "medium_sand": 6,
            "coarse_sand": 7,
            "gravel": 8,
            "shells": 9,
        }
        d = Lithology.to_dict(key="value")
        assert d == {
            0: "anthropogenic",
            1: "organic",
            2: "clay",
            3: "loam",
            5: "fine_sand",
            6: "medium_sand",
            7: "coarse_sand",
            8: "gravel",
            9: "shells",
        }
        with pytest.raises(ValueError):
            Lithology.to_dict(key="foo")  # Must raise ValueError with invalid input key



import warnings
from pathlib import Path
from typing import Any, Iterable, Literal

import geopandas as gpd
import pandas as pd
from pyproj import CRS

from geost import utils
from geost.base import (
    BoreholeCollection,
    Collection,
    CptCollection,
    DiscreteData,
    LayeredData,
    PointHeader,
)
from geost.bro import BroApi
from geost.io import _parse_cpt_gef_files, xml
from geost.io.parsers import BorisXML

type BroObject = Literal["BHR-GT", "BHR-P", "BHR-G", "CPT", "SFR"]

MANDATORY_LAYERED_DATA_COLUMNS = ["nr", "x", "y", "surface", "end", "top", "bottom"]
MANDATORY_DISCRETE_DATA_COLUMNS = ["nr", "x", "y", "surface", "end", "depth"]
MANDATORY_INCLINED_LAYERED_DATA_COLUMNS = [
    "nr",
    "x",
    "y",
    "x_bot",
    "y_bot",
    "surface",
    "end",
    "top",
    "bottom",
]
GEOMETRY_TO_SELECTION_FUNCTION = {
    "Polygon": "select_within_polygons",
    "Line": "select_with_lines",
    "Point": "select_with_points",
}
LLG_COLUMN_MAPPING = {
    "BOORP": "nr",
    "XCO": "x",
    "YCO": "y",
    "MV_HoogteNAP": "surface",
    "BoringEinddiepteNAP": "end",
    "BEGIN_DIEPTE": "top",
    "EIND_DIEPTE": "bottom",
}


def adjust_z_coordinates(data_dataframe: pd.DataFrame) -> pd.DataFrame:
    """
    Interpret data containing elevation or z-coordinates. GeoST data objects require
    that layer top/bottom or discrete data point z-coordinates to be increasing
    downward, starting at at 0. This function detects how elevation is currently defined
    and turns it into the desired format if required.

    Parameters
    ----------
    data_dataframe : pd.DataFrame
        Dataframe with layer or discrete data that includes columns referecing layer
        top, bottoms or discrete data z-coordinates.
    """
    top_column_label = [
        col for col in data_dataframe.columns if col in ["top", "depth"]
    ][0]
    has_bottom_column = "bottom" in data_dataframe.columns

    # TODO: think about detection. Only considers first two indices to determine
    # downward decreasing or increasing of z-coordinates.
    first_surface = data_dataframe["surface"].iloc[0]
    first_top = data_dataframe[top_column_label].iloc[0]
    second_top = data_dataframe[top_column_label].iloc[1]

    if first_top == first_surface:
        data_dataframe[top_column_label] -= data_dataframe["surface"]
        if has_bottom_column:
            data_dataframe["bottom"] -= data_dataframe["surface"]
    if first_top > second_top:
        data_dataframe[top_column_label] *= -1
        if has_bottom_column:
            data_dataframe["bottom"] *= -1

    return data_dataframe


def _check_mandatory_column_presence(
    df: pd.DataFrame, mandatory_cols: list, column_mapper: dict = None
) -> pd.DataFrame:
    missing_mandatory = [col for col in mandatory_cols if col not in df.columns]

    if missing_mandatory:
        if not column_mapper:
            column_mapper = dict()
            print(
                f"\nMandatory columns {missing_mandatory} are missing in the "
                f"data table. Columns present are: \n{list(df.columns)}\n"
            )
            for missing in missing_mandatory:
                name = input(f"Which column is {missing}? ")
                column_mapper[name] = missing

        df.rename(columns=column_mapper, inplace=True)
        df = df.loc[:, ~df.columns.duplicated(keep="last")]

    return df


def read_borehole_table(
    file: str | Path,
    horizontal_reference: str | int | CRS = 28992,
    vertical_reference: str | int | CRS = 5709,
    has_inclined: bool = False,
    as_collection: bool = True,
    column_mapper: dict = None,
    **kwargs,
) -> BoreholeCollection:
    """
    Read tabular borehole information. This is a file (parquet, csv or Excel) that
    includes row data for each (borehole) layer and must at least include the following
    column headers:

    - **nr** : Object id
    - **x** : X-coordinates according to the given horizontal_reference
    - **y** : Y-coordinates according to the given horizontal_reference
    - **surface** : Surface elevation according to the given vertical_reference
    - **end** : End depth according to the given vertical_reference
    - **top** : Top depth of each layer
    - **bottom** : Bottom depth of each layer

    In case there are inclined boreholes in the layer data, you additionally require:

    - **x_bot** : X-coordinates of the layer bottoms
    - **y_bot** : Y-coordinates of the layer bottoms

    If you are reading a file that uses different column names, see the optional
    argument "column_mapper" below. There are no further limits or requirements to
    additional (data) columns.

    Parameters
    ----------
    file : str | Path
        Path to file to be read. Depending on the file extension, the corresponding
        Pandas read function will be called. This can be either pandas.read_parquet,
        pandas.read_csv or pandas.read_excel. The **kwargs that can be given to this
        function will be passed to the selected Panda's read function.
    horizontal_reference : str | int | CRS, optional
        EPSG of the data's horizontal reference. Takes anything that can be interpreted
        by pyproj.crs.CRS.from_user_input(). The default is 28992.
    vertical_reference : str | int | CRS, optional
        EPSG of the data's vertical datum. Takes anything that can be interpreted by
        pyproj.crs.CRS.from_user_input(). However, it must be a vertical datum. FYI:
        "NAP" is EPSG 5709 and The Belgian reference system (Ostend height) is ESPG
        5710. The default is 5709.
    has_inclined : bool, optional
        If True, the borehole data table contains inclined boreholes. The default is False.
    as_collection : bool, optional
        If True, the borehole table will be read as a :class:`~geost.base.Collection`
        which includes a header object and spatial selection functionality. If False,
        a :class:`~geost.base.LayeredData` object is returned. The default is True.
    column_mapper: dict, optional
        If the file to be read uses different column names than the ones given in the
        description above, you can use a dictionary mapping to translate the column
        names to the required format. column_mapper is None by default, in which case
        the user is asked for input if not all required columns are encountered in the
        file.
    kwargs: optional
        Optional keyword arguments for Pandas.read_parquet, Pandas.read_csv or
        Pandas.read_excel.

    Returns
    -------
    :class:`~geost.base.BoreholeCollection` or :class:`~geost.base.LayeredData`
        Instance of :class:`~geost.base.BoreholeCollection` or :class:`~geost.base.LayeredData`
        depending on if the table is read as a collection or not.

    Examples
    --------
    Suppose we have a file of boreholes with columns 'nr', 'x', 'y', 'maaiveld', 'end',
    'top' and 'bottom'. The column 'maaiveld' corresponds to the required column
    'surface' in GeoST and thus needs to be remapped. The x and y-coordinates are given
    in WGS84 UTM 31N whereas the vertical datum is given in the Belgian Ostend height
    vertical datum:

    >>> read_borehole_table(file, 32631, 'Ostend height', column_mapper={'maaiveld': 'surface'})

    """
    boreholes = utils._pandas_read(file, **kwargs)

    if has_inclined:
        boreholes = _check_mandatory_column_presence(
            boreholes, MANDATORY_INCLINED_LAYERED_DATA_COLUMNS, column_mapper
        )
    else:
        boreholes = _check_mandatory_column_presence(
            boreholes, MANDATORY_LAYERED_DATA_COLUMNS, column_mapper
        )
    # Figure out top and bottom reference and coerce to downward increasing from 0 if
    # required.
    boreholes = adjust_z_coordinates(boreholes)
    boreholes = LayeredData(boreholes, has_inclined=has_inclined)

    if as_collection:
        boreholes = boreholes.to_collection(horizontal_reference, vertical_reference)

    return boreholes


def read_cpt_table(
    file: str | Path,
    horizontal_reference: str | int | CRS = 28992,
    vertical_reference: str | int | CRS = 5709,
    as_collection: bool = True,
    column_mapper: dict = None,
    **kwargs,
) -> CptCollection:
    """
    Read tabular CPT information. This is a file (parquet, csv or Excel) that includes
    row data for each CPT depth interval and must at least include the following header
    information as columns:

    - **nr** : Object id
    - **x** : X-coordinates according to the given horizontal_reference
    - **y** : Y-coordinates according to the given horizontal_reference
    - **surface** : Surface elevation according to the given vertical_reference
    - **end** : End depth according to the given vertical_reference
    - **depth** : Depth in meters below the surface

    Parameters
    ----------
    file : str | Path
        File with the CPT information to read.
    horizontal_reference : str | int | CRS, optional
        EPSG of the data's horizontal reference. Takes anything that can be interpreted
        by pyproj.crs.CRS.from_user_input(). The default is 28992.
    vertical_reference : str | int | CRS, optional
        EPSG of the data's vertical datum. Takes anything that can be interpreted by
        pyproj.crs.CRS.from_user_input(). However, it must be a vertical datum. FYI:
        "NAP" is EPSG 5709 and The Belgian reference system (Ostend height) is ESPG
        5710. The default is 5709.
    as_collection : bool, optional
        If True, the CPT table will be read as a :class:`~geost.base.Collection`
        which includes a header object and spatial selection functionality. If False,
        a :class:`~geost.base.DiscreteData` object is returned. The default is True.
    column_mapper: dict, optional
        If the file to be read uses different column names than the ones given in the
        description above, you can use a dictionary mapping to translate the column
        names to the required format. column_mapper is None by default, in which case
        the user is asked for input if not all required columns are encountered in the
        file.
    kwargs: optional
        Optional keyword arguments for Pandas.read_parquet, Pandas.read_csv or
        Pandas.read_excel.

    Returns
    -------
    :class:`~geost.base.CptCollection`
        Instance of :class:`~geost.base.CptCollection`.

    """
    cpts = utils._pandas_read(file, **kwargs)

    cpts = _check_mandatory_column_presence(
        cpts, MANDATORY_DISCRETE_DATA_COLUMNS, column_mapper
    )
    cpts = DiscreteData(cpts)

    if as_collection:
        cpts = cpts.to_collection(horizontal_reference, vertical_reference)

    return cpts


def read_nlog_cores(file: str | Path) -> BoreholeCollection:
    """
    Read NLog boreholes from the 'nlog_stratstelsel' Excel file. You can find this
    distribution of borehole data here: https://www.nlog.nl/boringen

    Warning: reading this Excel file is really slow (~10 minutes). Converting it to
    parquet using :func:`~geost.utils.excel_to_parquet` once and using that file instead
    allows for much faster reading of nlog data.

    Parameters
    ----------
    file : str | Path
        Path to nlog_stratstelsel.xlsx or .parquet

    Returns
    -------
    :class:`~geost.borehole.BoreholeCollection`
        :class:`~geost.borehole.BoreholeCollection`
    """
    nlog_cores = utils._pandas_read(file)

    nlog_cores.rename(
        columns={
            "NITG_NR": "nr",
            "X_TOP_RD": "x",
            "Y_TOP_RD": "y",
            "X_BOTTOM_RD": "x_bot",
            "Y_BOTTOM_RD": "y_bot",
            "TV_TOP_NAP": "top",
            "TV_BOTTOM_NAP": "bottom",
        },
        inplace=True,
    )

    nlog_cores["top"] *= -1
    nlog_cores["bottom"] *= -1

    nrs = []
    x = []
    y = []
    x_bot = []
    y_bot = []
    surface = []
    end = []
    nrs_data = []
    surface_data = []
    end_data = []
    for nr, data in nlog_cores.groupby("nr"):
        nrs.append(nr)
        x.append(data["x"].iloc[0])
        y.append(data["y"].iloc[0])
        x_bot.append(data["x_bot"].iloc[0])
        y_bot.append(data["y_bot"].iloc[0])
        surface.append(data["top"].iloc[0])
        end.append(data["bottom"].iloc[-1])
        nrs_data += [nr for i in range(len(data))]
        surface_data += [data["top"].iloc[0] for i in range(len(data))]
        end_data += [data["bottom"].iloc[-1] for i in range(len(data))]

    surface_end_df = pd.DataFrame(
        {"nr": nrs_data, "surface": surface_data, "end": end_data}
    ).drop_duplicates("nr")

    nlog_cores = nlog_cores.merge(surface_end_df, on="nr", how="left")
    nlog_cores = adjust_z_coordinates(nlog_cores)

    collection = LayeredData(nlog_cores, has_inclined=True).to_collection(28992, 5709)

    return collection


def read_xml_boris(
    file: str | Path,
    horizontal_reference: str | int | CRS = 28992,
    vertical_reference: str | int | CRS = 5709,
    as_collection: bool = True,
) -> BoreholeCollection:
    """
    Read export XML of the BORIS software. BORIS is software developed by TNO to
    describe boreholes in the lab. The exported XML-format bears no resemblance to DINO
    or BRO XML standards

    Parameters
    ----------
    file : str | Path
        File with the borehole information to read.
    horizontal_reference : str | int | CRS, optional
        EPSG of the data's horizontal reference. Takes anything that can be interpreted
        by pyproj.crs.CRS.from_user_input(). The default is 28992.
    vertical_reference : str | int | CRS, optional
        EPSG of the data's vertical datum. Takes anything that can be interpreted by
        pyproj.crs.CRS.from_user_input(). However, it must be a vertical datum. FYI:
        "NAP" is EPSG 5709 and The Belgian reference system (Ostend height) is ESPG
        5710. The default is 5709.
    as_collection : bool, optional
        If True, the CPT table will be read as a :class:`~geost.base.Collection`
        which includes a header object and spatial selection functionality. If False,
        a :class:`~geost.base.LayeredData` object is returned. The default is True.

    Returns
    -------
    :class:`~geost.base.BoreholeCollection` or :class:`~geost.base.LayeredData`
        Instance of :class:`~geost.base.BoreholeCollection` or :class:`~geost.base.LayeredData`
        depending on if the table is read as a collection or not.
    """
    boris_data = BorisXML(file)
    boreholes = LayeredData(boris_data.layer_dataframe, has_inclined=False)

    if as_collection:
        # Think of a better way to translate non-standard BORIS crs encoding or keep it
        # user-defined (like we do in other reader functions)
        boreholes = boreholes.to_collection(horizontal_reference, vertical_reference)

    return boreholes


def read_bhrgt(
    files: str | Path | Iterable[str | Path],
    company: str | None = None,
    schema: dict[str, Any] = None,
    read_all: bool = False,
) -> BoreholeCollection:
    """
    Read Geotechnical borehole (BHR-GT) XML file or files into a `BoreholeCollection`. This
    reads the XML files according to a `schema` that describes the structure of the of the
    XML data to find the relevant borehole data. Geost provides predefined schemas for the
    BHR-GT XML files that are used in the Netherlands from the Dutch national BRO database.
    A predefined schema may also be present for a specific company, which can be specified
    using the `company` parameter. Predefined schemas from Geost can be used and extended
    or you can provide your own schema to extract relevant attributes (see example link
    below).

    Parameters
    ----------
    files : str | Path | Iterable[str  |  Path]
        File or files to read.
    company : str | None, optional
        Company name to use a predefined schema for. See :mod:`geost.io.xml.schemas` for
        available schemas. If None, the predefined schema for BRO BHR-GT XML files is used
        by default to read the data. The default is None.
    schema : dict[str, Any], optional
        Schema to use for reading the data which describes the structure of the XML data.
        If not given and `company` is None, the predefined schema for BRO BHR-GT XML files
        is used.
    read_all : bool, optional
        XML files may contain multiple borehole elements per XML file. If True, the function
        will attempt to read all available data. Otherwise, only the first borehole element
        is read. The default is False.

    Returns
    -------
    :class:`~geost.base.BoreholeCollection`
        BoreholeCollection containing the header and data attributes of the BHR-GT data.

    Examples
    --------
    A full example of reading XML files can be found here:
    https://geost.readthedocs.io/en/latest/examples/bhrgt.html

    """
    header, data = xml.read(
        files, xml.read_bhrgt, company=company, schema=schema, read_all=read_all
    )

    header = PointHeader(
        gpd.GeoDataFrame(
            header, geometry=gpd.points_from_xy(header.x, header.y), crs=28992
        ),
        vertical_reference=5709,
    )
    data = LayeredData(data)

    return BoreholeCollection(header, data)


def read_bhrp(
    files: str | Path | Iterable[str | Path],
    company: str | None = None,
    schema: dict[str, Any] = None,
    read_all: bool = False,
) -> BoreholeCollection:
    """
    Read Pedological borehole (BHR-P) XML file or files into a `BoreholeCollection`. This
    reads the XML files according to a `schema` that describes the structure of the of the
    XML data to find the relevant borehole data. Geost provides predefined schemas for the
    BHR-P XML files that are used in the Netherlands from the Dutch national BRO database.
    A predefined schema may also be present for a specific company, which can be specified
    using the `company` parameter. Predefined schemas from Geost can be used and extended
    or you can provide your own schema to extract relevant attributes (see example link
    below).

    Parameters
    ----------
    files : str | Path | Iterable[str  |  Path]
        File or files to read.
    company : str | None, optional
        Company name to use a predefined schema for. See :mod:`geost.io.xml.schemas` for
        available schemas. If None, the predefined schema for BRO BHR-P XML files is used
        by default to read the data. The default is None.
    schema : dict[str, Any], optional
        Schema to use for reading the data which describes the structure of the XML data.
        If not given and `company` is None, the predefined schema for BRO BHR-P XML files
        is used.
    read_all : bool, optional
        XML files may contain multiple borehole elements per XML file. If True, the function
        will attempt to read all available data. Otherwise, only the first borehole element
        is read. The default is False.

    Returns
    -------
    :class:`~geost.base.BoreholeCollection`
        BoreholeCollection containing the header and data attributes of the BHR-P data.

    Examples
    --------
    A full example of reading XML files can be found here:
    https://geost.readthedocs.io/en/latest/examples/bhrp.html

    """
    header, data = xml.read(
        files, xml.read_bhrp, company=company, schema=schema, read_all=read_all
    )
    header = PointHeader(
        gpd.GeoDataFrame(
            header, geometry=gpd.points_from_xy(header.x, header.y), crs=28992
        ),
        vertical_reference=5709,
    )
    data = LayeredData(data)

    return BoreholeCollection(header, data)


def read_bhrg(
    files: str | Path | Iterable[str | Path],
    company: str | None = None,
    schema: dict[str, Any] = None,
    read_all: bool = False,
) -> BoreholeCollection:
    """
    Read Geological borehole (BHR-G) XML file or files into a `BoreholeCollection`. This
    reads the XML files according to a `schema` that describes the structure of the of the
    XML data to find the relevant borehole data. Geost provides predefined schemas for the
    BHR-G) XML file or files into a BoreholeCollection. This XML files that are used in the
    Netherlands from the Dutch national BRO database. A predefined schema may also be present
    for a specific company, which can be specified using the `company` parameter. Predefined
    schemas from Geost can be used and extended or you can provide your own schema to extract
    relevant attributes (see example link below).

    Parameters
    ----------
    files : str | Path | Iterable[str  |  Path]
        File or files to read.
    company : str | None, optional
        Company name to use a predefined schema for. See :mod:`geost.io.xml.schemas` for
        available schemas. If None, the predefined schema for BRO BHR-G) XML file or files
        into a BoreholeCollection. This XML files is used by default to read the data. The
        default is None.
    schema : dict[str, Any], optional
        Schema to use for reading the data which describes the structure of the XML data.
        If not given and `company` is None, the predefined schema for BRO BHR-G) XML file
        or files into a BoreholeCollection. This XML files is used.
    read_all : bool, optional
        XML files may contain multiple borehole elements per XML file. If True, the function
        will attempt to read all available data. Otherwise, only the first borehole element
        is read. The default is False.

    Returns
    -------
    :class:`~geost.base.BoreholeCollection`
        BoreholeCollection containing the header and data attributes of the BHR-G data.

    Examples
    --------
    A full example of reading XML files can be found here:
    https://geost.readthedocs.io/en/latest/examples/bhrgt.html

    """
    header, data = xml.read(
        files, xml.read_bhrg, company=company, schema=schema, read_all=read_all
    )
    header = PointHeader(
        gpd.GeoDataFrame(
            header, geometry=gpd.points_from_xy(header.x, header.y), crs=28992
        ),
        vertical_reference=5709,
    )
    data = LayeredData(data)

    return BoreholeCollection(header, data)


def read_sfr(
    files: str | Path | Iterable[str | Path],
    company: str | None = None,
    schema: dict[str, Any] = None,
    read_all: bool = False,
) -> BoreholeCollection:
    """
    Read Pedological soilprofile description (SFR) XML file or files into a `BoreholeCollection`.
    This reads the XML files according to a `schema` that describes the structure of the of
    the XML data to find the relevant borehole data. Geost provides predefined schemas for
    the SFR XML file or files into a BoreholeCollection. This XML files that are used in the
    Netherlands from the Dutch national BRO database. A predefined schema may also be present
    for a specific company, which can be specified using the `company` parameter. Predefined
    schemas from Geost can be used and extended or you can provide your own schema to extract
    relevant attributes (see example link below).

    Parameters
    ----------
    files : str | Path | Iterable[str  |  Path]
        File or files to read.
    company : str | None, optional
        Company name to use a predefined schema for. See :mod:`geost.io.xml.schemas` for
        available schemas. If None, the predefined schema for BRO SFR XML file or files
        into a BoreholeCollection. This XML files is used by default to read the data. The
        default is None.
    schema : dict[str, Any], optional
        Schema to use for reading the data which describes the structure of the XML data.
        If not given and `company` is None, the predefined schema for BRO SFR XML file or
        files into a BoreholeCollection. This XML files is used.
    read_all : bool, optional
        XML files may contain multiple borehole elements per XML file. If True, the function
        will attempt to read all available data. Otherwise, only the first borehole element
        is read. The default is False.

    Returns
    -------
    :class:`~geost.base.BoreholeCollection`
        BoreholeCollection containing the header and data attributes of the SFR data.

    Examples
    --------
    A full example of reading XML files can be found here:
    https://geost.readthedocs.io/en/latest/examples/bhrgt.html

    """
    header, data = xml.read(
        files, xml.read_sfr, company=company, schema=schema, read_all=read_all
    )
    header = PointHeader(
        gpd.GeoDataFrame(
            header, geometry=gpd.points_from_xy(header.x, header.y), crs=28992
        ),
        vertical_reference=5709,
    )
    data = LayeredData(data)

    return BoreholeCollection(header, data)


def read_gef_cpts(file_or_folder: str | Path) -> CptCollection:
    """
    Read one or more GEF files of CPT data into a geost CptCollection.

    Parameters
    ----------
    file_or_folder : str | Path
        GEF files to read.

    Returns
    -------
    :class:`~geost.base.CptCollection`
        :class:`~geost.base.CptCollection` of the GEF file(s).

    """
    data = _parse_cpt_gef_files(file_or_folder)
    df = pd.concat(data)

    return DiscreteData(df).to_collection()


def read_cpt(
    files: str | Path | Iterable[str | Path],
    company: str | None = None,
    schema: dict[str, Any] = None,
    read_all: bool = False,
) -> CptCollection:
    """
    Read Cone Penetration Test (CPT) XML file or files into a `CptCollection`. This reads
    the XML files according to a `schema` that describes the structure of the of the XML
    data to find the relevant CPT data. Geost provides predefined schemas for the CPT XML
    file or files into a CptCollection. This XML files that are used in the Netherlands
    from the Dutch national BRO database. A predefined schema may also be present for a
    specific company, which can be specified using the `company` parameter. Predefined
    schemas from Geost can be used and extended or you can provide your own schema to
    extract relevant attributes (see example link below).

    Parameters
    ----------
    files : str | Path | Iterable[str  |  Path]
        File or files to read.
    company : str | None, optional
        Company name to use a predefined schema for. See :mod:`geost.io.xml.schemas` for
        available schemas. If None, the predefined schema for BRO CPT XML file or files
        into a CptCollection. This XML files is used by default to read the data. The
        default is None.
    schema : dict[str, Any], optional
        Schema to use for reading the data which describes the structure of the XML data.
        If not given and `company` is None, the predefined schema for BRO CPT XML file or
        files into a CptCollection. This XML files is used.
    read_all : bool, optional
        XML files may contain multiple CPT elements per XML file. If True, the function
        will attempt to read all available data. Otherwise, only the first CPT element
        is read. The default is False.

    Returns
    -------
    :class:`~geost.base.CptCollection`
        CptCollection containing the header and data attributes of the CPT data.

    Examples
    --------
    A full example of reading XML files can be found here:
    https://geost.readthedocs.io/en/latest/examples/bhrgt.html

    """
    header, data = xml.read(
        files, xml.read_cpt, company=company, schema=schema, read_all=read_all
    )
    header = PointHeader(
        gpd.GeoDataFrame(
            header, geometry=gpd.points_from_xy(header.x, header.y), crs=28992
        ),
        vertical_reference=5709,
    )
    data.fillna({"depth": data["penetrationlength"]}, inplace=True)
    data = DiscreteData(data)

    return CptCollection(header, data)


def read_uullg_tables(
    header_table: str | Path,
    data_table: str | Path,
    horizontal_reference: str | int | CRS = 28992,
    vertical_reference: str | int | CRS = 5709,
    **kwargs,
) -> BoreholeCollection:
    """
    Read the header and data tables from UU-LLG boreholes (see: EasyDans KNAW) from a
    local csv or parquet file into a GeoST BoreholeCollection object.

    Reference dataset: https://easy.dans.knaw.nl/ui/datasets/id/easy-dataset:74935
    See https://wikiwfs.geo.uu.nl/ for additional information.

    Parameters
    ----------
    header_table : str | Path
        Path to file of the UU-LLG header table.
    data_table : str | Path
        Path to file of the UU-LLG data table.
    horizontal_reference : str | int | CRS, optional
        EPSG of the data's horizontal reference. Takes anything that can be interpreted
        by pyproj.crs.CRS.from_user_input(). The default is 28992.
    vertical_reference : str | int | CRS, optional
        EPSG of the data's vertical datum. Takes anything that can be interpreted by
        pyproj.crs.CRS.from_user_input(). However, it must be a vertical datum. FYI:
        "NAP" is EPSG 5709 and The Belgian reference system (Ostend height) is ESPG
        5710. The default is 5709.

    Returns
    -------
    :class:`~geost.base.BoreholeCollection`
        Instance of :class:`~geost.base.BoreholeCollection` of the UU-LLG data.

    """
    header = utils._pandas_read(header_table, **kwargs)
    data = utils._pandas_read(data_table, **kwargs)

    header.rename(columns=LLG_COLUMN_MAPPING, inplace=True)
    data.rename(columns=LLG_COLUMN_MAPPING, inplace=True)

    header = header.astype({"nr": str})
    data = data.astype({"nr": str})

    header = utils.dataframe_to_geodataframe(header).set_crs(horizontal_reference)

    add_header_cols = [c for c in ["x", "y", "surface", "end"] if c not in data.columns]
    if add_header_cols:
        data = data.merge(header[["nr"] + add_header_cols], on="nr", how="left")

    header = PointHeader(header, vertical_reference)
    data = LayeredData(data)
    return BoreholeCollection(header, data)


def read_collection_geopackage(
    filepath: str | Path,
    collection_type: Collection,
    horizontal_reference: str | int | CRS = 28992,
    vertical_reference: str | int | CRS = 5709,
):
    """
    Read a GeoST stored Geopackage file of `geost.base.Collection` objects. The
    Geopackage contains the Collection's "header" and "data" attributes as layers which
    are read into the specified Collection.

    Parameters
    ----------
    filepath : str | Path
        GeoST stored Geopackage file of a `geost.base.Collection` object.
    collection_type : Collection
        Type of GeoST Collection object the data needs to be. Subclasses of `Collection`
        (e.g. :class:`geost.base.BoreholeCollection`, :class:`geost.base.CptCollection`)
        for the available types.
    horizontal_reference : str | int | CRS, optional
        EPSG of the data's horizontal reference. Takes anything that can be interpreted
        by pyproj.crs.CRS.from_user_input(). The default is 28992.
    vertical_reference : str | int | CRS, optional
        EPSG of the data's vertical datum. Takes anything that can be interpreted by
        pyproj.crs.CRS.from_user_input(). However, it must be a vertical datum. FYI:
        "NAP" is EPSG 5709 and The Belgian reference system (Ostend height) is ESPG
        5710. The default is 5709.

    Returns
    -------
    Subclass of `geost.base.Collection`
        Subclass which is specified as collection_type.

    Raises
    ------
    ValueError
        Raises ValueError if the collection_type is not supported.

    Examples
    --------
    Any GeoST `Collection` object can be stored as a geopackage file:

    >>> original_collection = geost.data.boreholes_usp()
    >>> original_collection
    BoreholeCollection:
    # header = 67
    >>> original_collection.to_geopackage("./collection.gpkg")

    Reading the stored collection using `geost.read_collection_geopackage` returns the
    stored collection:

    >>> collection = geost.read_collection_geopackage(
    ...     "./collection.gpkg", collection_type=BoreholeCollection
    ...     )
    >>> collection
    BoreholeCollection:
    # header = 67

    """
    # TODO: Check collection type inference possibility from geopackage metadata.
    if collection_type == BoreholeCollection:
        header_type = PointHeader
        data_type = LayeredData
    elif collection_type == CptCollection:
        header_type = PointHeader
        data_type = DiscreteData
    else:
        raise ValueError(f"Collection type {collection_type} not supported.")

    header = gpd.read_file(filepath, layer="header")
    header.set_crs(horizontal_reference, inplace=True, allow_override=True)
    data = gpd.read_file(filepath, layer="data")
    return collection_type(header_type(header, vertical_reference), data_type(data))


def read_pickle(filepath: str | Path, **kwargs) -> Any:
    """
    Read pickled GeoST object (or any object) from file.

    Parameters
    ----------
    filepath : str | Path
        Path to pickle file to open.
    **kwargs
        pd.read_pickle kwargs. See relevant Pandas documentation.

    Returns
    -------
    Any
        Returns the Python object that was stored in the pickle.

    Examples
    --------
    Any GeoST :class:`geost.base.Collection` object can be stored as a pickle file:

    >>> original_collection = geost.data.boreholes_usp()
    >>> original_collection
    BoreholeCollection:
    # header = 67
    >>> original_collection.to_pickle("./collection.pkl")

    Reading the stored pickle using `geost.read_pickle` returns the stored collection:

    >>> collection = geost.read_pickle("./collection.pkl")
    >>> collection
    BoreholeCollection:
    # header = 67

    """
    pkl = pd.read_pickle(filepath, **kwargs)
    return pkl


def bro_api_read(
    object_type: BroObject,
    *,
    bro_ids: str | list[str] = None,
    bbox: tuple[float, float, float, float] = None,
    geometry: gpd.GeoDataFrame = None,
    buffer: int | float = None,
    schema: dict[str, Any] = None,
):
    """
    Read data directly from the BRO API. This allows to read BHR-GT, BHR-P, BHR-G and
    CPT data from the BRO API into `geost` objects without having to download the data
    first. Note, API requests are relatively slow, so this function is not meant for bulk
    downloads. Also, the API is limited to 2000 objects per request, so if you request
    more than 2000 objects, the API will return an error.

    Parameters
    ----------
    object_type : BroObject
        Type of BRO object to read. This can be one of the following: "BHR-GT", "BHR-P",
        "BHR-G", "CPT", "SFR".
    bro_ids : str | list[str], optional (keyword only)
        List of BRO object ids to read. If not given, the bbox or geometry must be
        provided to select the objects to read. If given, bbox and geometry are ignored.
    bbox : tuple[float, float, float, float], optional
        Bounding box (xmin, ymin, xmax, ymax) to search BRO objects within. The default is
        None.
    geometry : gpd.GeoDataFrame, optional (keyword only)
        Geometry to search within. Can be any Shapely geometry object (Point, Line, Polygon),
        Geopandas GeoDataFrame of path to a geometry file. The function retrieve the BRO objects
        that are within the outermost bounds (i.e. bounding box) of the input geometry. The
        default is None.
    buffer : int | float, optional (keyword only)
        Buffer distance to apply to the geometry. This increases the bounding box of the the
        input geometry. The default is None.
    schema : dict[str, Any], optional (keyword only)
        XML schema to use for reading the data which describes the structure of the XML
        data. If not given, the function will use the default schema for the specified
        object_type provided by Geost (see `geost.io.xml.schemas`). Note that predefined
        schemas only retrieve the attributes that are defined in the schema but more data
        may be available in the requested objects. To retrieve desired attributes, you can
        modify the schema accordingly (see usage examples).

    Returns
    -------
    Subclass of `geost.base.Collection`
        Returns a subclass of `geost.base.Collection` depending on the object_type.

    Examples
    --------
    Read a list of specific borehole ids from the BRO API (max 2000 objects):
    >>> import geost
    ... bhrgt = geost.bro_api_read("BHR-GT", bro_ids=["BHR000000339725", "BHR000000339735"])
    ... bhrgt
    BoreholeCollection:
    # header = 2

    Or read geological boreholes (BHR-G) within a bounding box:
    >>> import geost
    ... bbox = (126_800, 448_000, 127_800, 449_000)  # xmin, ymin, xmax, ymax
    ... bhrg = geost.bro_api_read("BHR-G", bbox=bbox)
    BoreholeCollection:
    # header = 3

    Or within shapefile containing a Polygon geometry of the same bounding box of the
    previous example (Point or Line geometries are also supported):
    >>> import geost
    ... bhrg = geost.bro_api_read("BHR-G", geometry="my_polygon.shp")
    BoreholeCollection:
    # header = 3

    Or within the shapefile and applying a buffer:
    >>> import geost
    ... bhrg = geost.bro_api_read("BHR-G", geometry="my_polygon.shp", buffer=500)
    BoreholeCollection:
    # header = 5

    When there are no objects found, you will get an empty collection:
    >>> import geost
    ... bhrg = geost.bro_api_read("BHR-G", bbox=[0, 0, 1, 1])
    Collection:
    <EMPTY COLLECTION>
    """
    readers = {
        "BHR-GT": read_bhrgt,
        "BHR-P": read_bhrp,
        "BHR-G": read_bhrg,
        "CPT": read_cpt,
        "SFR": read_sfr,
    }
    reader = readers.get(object_type, None)
    if reader is None:
        raise ValueError(f"Object type '{object_type}' is not supported for reading.")

    api = BroApi()

    if bro_ids is not None:
        bro_data = api.get_objects(bro_ids, object_type=object_type)
    else:
        if bbox is not None:
            xmin, ymin, xmax, ymax = bbox
        elif geometry is not None:
            geometry = utils.check_geometry_instance(geometry)
            if buffer is not None:
                geometry = geometry.buffer(buffer)
            xmin, ymin, xmax, ymax = geometry.total_bounds
        api.search_objects_in_bbox(xmin, ymin, xmax, ymax, object_type=object_type)
        bro_data = api.get_objects(api.object_list, object_type=object_type)

    if bro_data:
        collection = reader(bro_data, schema=schema)
    else:
        collection = Collection(None, None)

    return collection
